import os
import tempfile

from packy_agent.configuration.settings import Settings
from packy_agent.configuration.sources.base import MappingBackedSource


def test_settings_from_file():
    with tempfile.NamedTemporaryFile(buffering=0) as f:
        f.write(b'value1: 10\nvalue2: 15')

        os.environ['PACKY_AGENT_SETTINGS_FILENAME'] = f.name

        settings = Settings()

        assert settings.get_value1() == 10
        assert settings.get_value2() == 15


def test_settings_from_command_line():
    settings = Settings()
    settings.set_commandline_arguments({'value1': 10, 'value2': 15})

    assert settings.get_value1() == 10
    assert settings.get_value2() == 15


def test_runtime_settings():
    settings = Settings()
    assert settings.get_component() is None

    settings.set_runtime('component', 'test')

    assert settings.get_component() == 'test'


def test_nested_dicts():
    settings = Settings()

    settings._sources = (
        (MappingBackedSource(
            {
                '1': {'1.2': {
                        '1.2.2': 'new 122',  # replace value
                        '1.2.3': 'new 123',  # add new
                    }
                },
                '2': {'2.1': 'different type'},  # replace with different type
                '3': 'new 3',  # add new at root level
            }, name='priority 1'
        )),
        (MappingBackedSource(
            {
                '1': {
                    '1.1': {'1.1.1': '111'},
                    '1.2': {
                        '1.2.1': '121',
                        '1.2.2': '122',
                    }
                },
                '2': {
                    '2.1': {'2.1.1': '211'},
                    '2.2': {
                        '2.2.1': '221',
                        '2.2.2': '222',
                    }
                }
            }, name='priority 2'
        )),
    )

    value1 = settings.get_1()
    assert value1['1.2']['1.2.1'] == '121'
    assert value1['1.2']['1.2.2'] == 'new 122'
    assert value1['1.2']['1.2.3'] == 'new 123'

    value2 = settings.get_2()
    assert value2['2.1'] == 'different type'
    assert value2['2.2'] == {'2.2.1': '221', '2.2.2': '222'}

    value3 = settings.get_3()
    assert value3 == 'new 3'


def test_get_set_section():
    from packy_agent.configuration.settings import settings
    assert settings.get_console_test_key() is None
    settings.set_console_test_key('test_value')
    assert settings.get_console_test_key() == 'test_value'
    console = settings.get_console()
    assert 'test_key' in console
    assert console['test_key'] == 'test_value'



import logging
from io import BytesIO

import supervisor.supervisorctl
from supervisor.options import ClientOptions

from packy_agent.configuration.settings import settings
from packy_agent.utils.misc import get_executable_path
from packy_agent.utils.shell import run_command_async

logger = logging.getLogger(__name__)


class CustomController(supervisor.supervisorctl.Controller):

    def output(self, message):
        if isinstance(message, str):
            message = message.encode('utf-8')
        self.stdout.write(message + b'\n')


def supervisor_main(args=None, options=None, stdout=None):
    if options is None:
        options = ClientOptions()

    options.realize(args, doc=__doc__)
    c = CustomController(options, stdout=stdout)

    if options.args:
        c.onecmd(' '.join(options.args))

    if options.interactive:
        c.exec_cmdloop(args, options)


def run_supervisor_command_sync(command, *args):
    logger.debug('Running sync supervisor command: %s %s', command, args)
    stdout = BytesIO()
    supervisor_main(['-c', settings.get_supervisor_configuration_filename(), command] + list(args),
                    stdout=stdout)

    stdout.seek(0)
    return stdout.read().decode('utf-8')


def run_supervisor_command_async(command, args, delay_seconds=None, dev_null=False):
    logger.debug('Running async supervisor command: %s %s', command, args)
    shell_command = '{executable} -c {configuration_filename} {command} {args}'.format(
        executable=get_executable_path('supervisorctl'),
        configuration_filename=settings.get_supervisor_configuration_filename(),
        command=command,
        args=' '.join(args))
    run_command_async(shell_command, delay_seconds, dev_null=dev_null)


def run_supervisor_command(command, args, is_async=False, async_delay_seconds=None,
                           async_dev_null=False):
    if is_async:
        run_supervisor_command_async(command, args, delay_seconds=async_delay_seconds,
                                     dev_null=async_dev_null)
    else:
        return run_supervisor_command_sync(command, *args)



import time

from sqlalchemy import Column, BigInteger, Integer, BLOB, UnicodeText

from packy_agent.domain_logic.models.sqlalchemy.base import ModelBase


class Measurement(ModelBase):
    __tablename__ = 'measurement'

    id = Column(BigInteger().with_variant(Integer, 'sqlite'), primary_key=True, autoincrement=True)
    measurement_type = Column(Integer, nullable=False)
    created_at_ts = Column(Integer, nullable=False, default=time.time)
    submitted_at_ts = Column(Integer)
    error_at_ts = Column(Integer)
    error_message = Column(UnicodeText)
    error_side = Column(Integer)
    value = Column(BLOB, nullable=False)



from flask_wtf import FlaskForm
from wtforms import SelectField, HiddenField


class ActivateExtraForm(FlaskForm):

    email = HiddenField()
    password = HiddenField()
    extra = HiddenField()

    agent = SelectField('Activate as', choices=(('new', 'New agent'),))



import argparse
import logging
import sys
import time
import socket

from packy_agent.configuration.settings import settings
from packy_agent.utils.cli import get_base_argument_parser
from packy_agent.utils.logging import configure_logging, configure_logging_basic

from packy_agent.domain_logic.modules.ping import base


def do_ping(host, timeout=2, ping_count=None, interval_seconds=1, packet_size_bytes=56):
    try:
        destination_ip_address = socket.gethostbyname(host)
    except socket.gaierror:
        print('Could not resolve {} to IP address'.format(host))
        return

    print('PING {} ({})'.format(host, destination_ip_address))
    no = 1
    while ping_count is None or ping_count > 0:
        rtt_seconds = base.ping_once(host, timeout=timeout, packet_size_bytes=packet_size_bytes)
        print('{0} time={1:.2f} ms'.format(no, round(rtt_seconds * 1000, 2)))

        no += 1
        if ping_count is not None:
            ping_count -= 1

        time.sleep(interval_seconds)


def entry():
    configure_logging_basic(logging.WARNING)
    parser = get_base_argument_parser(__loader__.name)

    parser.add_argument('host')
    parser.add_argument('--ping-count', type=int, default=None)
    parser.add_argument('--packet-size-bytes', help='ICMP payload size in bytes',
                        type=int, default=56)
    parser.add_argument('--interval-seconds', type=float, default=1)
    parser.add_argument('--timeout', type=float, default=2)

    args = parser.parse_args()
    settings.set_commandline_arguments(vars(args))
    configure_logging(logging.WARNING)

    return do_ping(args.host, timeout=args.timeout, ping_count=args.ping_count,
                   interval_seconds=args.interval_seconds,
                   packet_size_bytes=args.packet_size_bytes)


if __name__ == '__main__':
    sys.exit(entry())



import json

import httpretty

from packy_agent.configuration.settings import settings


def test_reactivation(console_flask_client):
    assert not settings.is_activated()

    response = console_flask_client.get('/activate/')
    assert response.status_code == 200

    with httpretty.enabled():
        url = settings.get_server_base_url() + 'api/v2/user/agent/'
        httpretty.register_uri(httpretty.GET, url, body='[{"id": 1, "name": "deactivated"}]')
        httpretty.register_uri(
            httpretty.PATCH, url + '1/', status=200,
            body=json.dumps({
                'client_id': 'client_id',
                'access_token': 'access_token',
                'refresh_token': 'refresh_token',
                'key': 'key',
                'name': 'name',
            }))

        response = console_flask_client.post(
            '/activate/', data={'email': 'test@test.com', 'password': 'test'},
            follow_redirects=True)
        assert response.status_code == 200
        response = console_flask_client.post(
            '/activate/', data={'email': 'test@test.com', 'password': 'test', 'agent': '1',
                                'extra': 'yes'},
            follow_redirects=True)
        assert response.status_code == 200

        latest_requests = httpretty.HTTPretty.latest_requests
        assert len(latest_requests) >= 3
        assert latest_requests[0].method == 'GET'
        assert latest_requests[0].path == '/api/v2/user/agent/?fields=id%2Cname&is_active=false'
        assert latest_requests[1].method == 'GET'
        assert latest_requests[1].path == '/api/v2/user/agent/?fields=id%2Cname&is_active=false'
        assert latest_requests[2].method == 'PATCH'
        assert latest_requests[2].path == '/api/v2/user/agent/1/?with_configuration=yes'
        body = json.loads(latest_requests[2].body)
        assert 'is_active' in body
        assert body['is_active'] is True

    assert settings.is_activated()



import sys
import logging

import yaml
import termcolor

from packy_agent.configuration.settings import settings
from packy_agent.utils.logging import configure_logging, configure_logging_basic
from packy_agent.utils.cli import get_base_argument_parser

LIST_COMMAND = 'list'
SET_COMMAND = 'set'
GET_COMMAND = 'get'
DEL_COMMAND = 'del'
CLEAR_COMMAND = 'clear'

is_a_tty = sys.stdout.isatty()


def print_key_value(key, value, is_shadowed=False):
    yaml_value = yaml.dump(value)
    if yaml_value.endswith('...\n'):
        yaml_value = yaml_value[:-4]

    yaml_value = yaml_value.rstrip('\n')

    text = '{}={}'.format(key, yaml_value)
    if is_shadowed:
        if is_a_tty:
            termcolor.cprint(text, attrs=('dark',))
        else:
            print(f'/ {text} /')
    else:
        print(text)


def list(args):
    all_keys = not getattr(args, 'effective', False)
    known_keys = set()
    for priority, (label, items) in enumerate(settings.labeled_items(), start=1):
        header = f'PRIORITY {priority}: {label}'
        print(header)
        print('=' * len(header))
        for key, value in items:
            is_shadowed = key in known_keys
            if all_keys or not is_shadowed:
                print_key_value(key, value, is_shadowed=is_shadowed)

            known_keys.add(key)

        print()


def set_key(args):
    settings.set(args.key, args.value)


def del_key(args):
    settings.delete(args.key)


def clear(args):
    settings.clear()


def get_value(args):
    key = args.key
    print_key_value(key, settings.get(key, '<NOT SET>'))


COMMAND_TO_FUNCTION_MAP = {
    LIST_COMMAND: list,
    SET_COMMAND: set_key,
    GET_COMMAND: get_value,
    DEL_COMMAND: del_key,
    CLEAR_COMMAND: clear,
}


def entry():
    configure_logging_basic(logging.WARNING)
    parser = get_base_argument_parser(__loader__.name)

    subparsers = parser.add_subparsers(title='command', dest='command')
    list_parser = subparsers.add_parser(LIST_COMMAND, help='List all known settings')
    list_parser.add_argument('-e', '--effective', action='store_true')
    get_parser = subparsers.add_parser(GET_COMMAND, help='Get particular setting')
    get_parser.add_argument('key')
    set_parser = subparsers.add_parser(SET_COMMAND, help='Set particular setting (local)')
    set_parser.add_argument('key')
    set_parser.add_argument('value')
    set_parser = subparsers.add_parser(DEL_COMMAND, help='Delete particular setting (local)')
    set_parser.add_argument('key')
    subparsers.add_parser(CLEAR_COMMAND, help='Clear settings (local)')

    args = parser.parse_args()

    settings.set_commandline_arguments(vars(args))
    configure_logging(logging.WARNING)

    # Call function named `args.command` and pass `args` to it
    COMMAND_TO_FUNCTION_MAP[args.command or LIST_COMMAND](args)


if __name__ == '__main__':
    sys.exit(entry())



from urllib.parse import urlencode

import flask
from werkzeug.routing import BuildError


def get_url(endpoint):
    try:
        return flask.url_for(endpoint)
    except BuildError:
        return endpoint


def smart_redirect(endpoint, back_endpoint, button_text='Back'):
    back = urlencode({
        'back_url': get_url(back_endpoint),
        'button_text': button_text,
    })
    url = get_url(endpoint) + '?' + back
    return flask.redirect(url)



import logging

import sentry_sdk
from sentry_sdk.integrations.logging import LoggingIntegration
from sentry_sdk.integrations.flask import FlaskIntegration

import packy_agent
from packy_agent.utils.platforms import get_platform
from packy_agent.configuration.settings import settings

logger = logging.getLogger(__name__)


def init_sentry_client(flask_integration=False, packy_agent_sentry_dsn=None, **kwargs):
    if not packy_agent_sentry_dsn:
        packy_agent_sentry_dsn = settings.get_sentry_dsn()

    if packy_agent_sentry_dsn:
        component = settings.get_component()
        platform_name = get_platform()
        release = f'Packy Agent v{packy_agent.__version__} / {component} / {platform_name}'

        integrations = [LoggingIntegration(level=logging.ERROR, event_level=logging.ERROR)]
        if flask_integration:
            integrations.append(FlaskIntegration())

        agent_key = settings.get_agent_key() or ''
        agent_name = settings.get_agent_name() or ''
        kwargs_ = {
            'debug': True,
            'release': release,
            'environment': settings.get_server_base_url() or '',
            'server_name': f'{agent_name} [{agent_key}]',
            'request_bodies': 'medium',
            'integrations': integrations,
        }
        kwargs_.update(kwargs)

        sentry_sdk.init(packy_agent_sentry_dsn, **kwargs_)
    else:
        logger.warning('Sentry DSN is not configured')



import os
from collections import OrderedDict

from packy_agent.configuration.sources.base import MappingBackedSource
from packy_agent.configuration.base import NESTING_PREFIXES
from packy_agent.utils.misc import yaml_coerce
from packy_agent.utils.collections import deep_update


ENVVAR_PREFIX = 'PACKY_AGENT_'
ENVVAR_PREFIX_LEN = len(ENVVAR_PREFIX)
NESTING_PREFIXES_ = tuple(v.upper() + '_' for v in NESTING_PREFIXES)


def get_settings_key_value(env_key, env_value):
    coerce_value = yaml_coerce(env_value)
    level1_key = env_key[ENVVAR_PREFIX_LEN:]
    for nesting_prefix in NESTING_PREFIXES_:
        if level1_key.startswith(nesting_prefix):
            return {
                nesting_prefix[:-1].lower(): {
                    level1_key[len(nesting_prefix):].lower(): coerce_value
                }
            }

    return {level1_key.lower(): coerce_value}


def is_settings_env_var(env_key):
    return env_key.startswith(ENVVAR_PREFIX)


class Environment(MappingBackedSource):

    def __init__(self, name=None):
        super().__init__(os.environ, name=name)

    def __getitem__(self, item):
        # TODO(dmu) LOW: Provide a faster implementation
        return dict(self.items())[item]

    def __iter__(self):
        yield from self.keys()

    def __contains__(self, item):
        # TODO(dmu) LOW: Provide a faster implementation
        return item in self.keys()

    def keys(self):
        yield from (k for k, _ in self.items())

    def items(self):
        all_items = OrderedDict()
        for key, value in super().items():
            if is_settings_env_var(key):
                deep_update(all_items, get_settings_key_value(key, value))

        yield from all_items.items()

    def values(self):
        yield from (v for _, v in self.items())

    def get(self, key, default=None):
        if not isinstance(key, str):
            raise TypeError("'key' must be a string")

        # TODO(dmu) LOW: Provide a faster implementation
        return dict(self.items()).get(key, default)



import logging

import sentry_sdk


logger = logging.getLogger(__name__)


def test_message_is_sent_to_sentry_when_message_is_sent_explicitly(sentry_server_mock):
    events = sentry_server_mock
    events.clear()
    sentry_sdk.capture_message('Test message')
    assert len(events) == 1
    first_event = events[0]
    assert first_event['message'] == 'Test message'


def test_message_is_sent_to_sentry_when_error_is_logged(sentry_server_mock):
    events = sentry_server_mock
    events.clear()
    logger.error('Test error')
    assert len(events) == 1
    first_event = events[0]
    assert first_event['logentry']['message'] == 'Test error'



class Skip(Exception):
    pass


class ImproperlyConfiguredError(Exception):
    pass


class AuthenticationError(Exception):
    pass


class ValidationError(Exception):
    pass


class CoolDown(Exception):
    pass


class NoAccessTokenError(ImproperlyConfiguredError):
    pass


class OutOfPortsError(Exception):
    pass



import os

import pytest

from packy_agent.configuration.sources.environment import Environment


def test_environment_smoke():
    os.environ['PACKY_AGENT_SMOKE'] = 'smoke test'
    conf = Environment()
    assert 'smoke' in conf
    assert conf['smoke'] == 'smoke test'


def test_mapping_backed_source():
    for key in os.environ:
        if key.startswith('PACKY_AGENT_'):
            del os.environ[key]

    os.environ['PACKY_AGENT_STRING'] = 'test'
    os.environ['PACKY_AGENT_INT'] = '10'
    os.environ['PACKY_AGENT_FLOAT'] = '1.5'
    os.environ['PACKY_AGENT_BOOLEAN'] = 'true'

    source = Environment()
    assert set(source) == {'string', 'int', 'float', 'boolean'}
    assert source['string'] == 'test'
    assert source['int'] == 10
    assert source['float'] == 1.5
    assert source['boolean']
    assert 'string' in source
    assert source.get('string') == 'test'
    assert source.get('something', 4) == 4
    assert sorted(source.keys()) == sorted(('string', 'int', 'float', 'boolean'))
    source.values()  # smoke
    assert sorted(source.items()) == [('boolean', True), ('float', 1.5), ('int', 10),
                                      ('string', 'test')]

    with pytest.raises(TypeError) as ex:
        source['value1'] = 0

    assert "'Environment' object does not support item assignment" in str(ex.value)

    with pytest.raises(TypeError) as ex:
        del source['value1']
    assert "'Environment' object does not support item deletion" in str(ex.value)

    with pytest.raises(AttributeError) as ex:
        source.pop('value1')
    assert "'Environment' object has no attribute 'pop'" in str(ex.value)


def test_environment_getitem_coercion():
    os.environ['PACKY_AGENT_STRING'] = 'test'
    os.environ['PACKY_AGENT_INT'] = '10'
    os.environ['PACKY_AGENT_FLOAT'] = '1.5'
    os.environ['PACKY_AGENT_BOOLEAN'] = 'true'
    os.environ['PACKY_AGENT_DICT'] = '{2: 3}'
    conf = Environment()
    assert 'string' in conf
    assert conf['string'] == 'test'
    assert 'int' in conf
    assert conf['int'] == 10
    assert isinstance(conf['int'], int)
    assert 'float' in conf
    assert conf['float'] == 1.5
    assert 'boolean' in conf
    assert conf['boolean']
    assert isinstance(conf['boolean'], bool)
    assert 'dict' in conf
    assert conf['dict'] == {2: 3}
    assert isinstance(conf['dict'], dict)


def test_environment_supports_nestings():
    os.environ['PACKY_AGENT_CONSOLE_HTTP_PORT'] = '9000'
    os.environ['PACKY_AGENT_SUPERVISOR_A'] = '1'
    os.environ['PACKY_AGENT_WORKER_B'] = '2'
    os.environ['PACKY_AGENT_WATCHDOG_C'] = '3'
    conf = Environment()
    assert 'console' in conf
    assert conf['console'] == {'http_port': 9000}
    assert 'supervisor' in conf
    assert conf['supervisor'] == {'a': 1}
    assert 'worker' in conf
    assert conf['worker'] == {'b': 2}
    assert 'watchdog' in conf
    assert conf['watchdog'] == {'c': 3}



import sys
import signal
import logging

from packy_agent.domain_logic.constants import WATCHDOG_COMPONENT
from packy_agent.watchdog.service import WatchdogService
from packy_agent.configuration.settings import settings
from packy_agent.utils.logging import configure_logging, configure_logging_basic
from packy_agent.utils.cli import get_base_argument_parser
from packy_agent.clients.sentry import init_sentry_client
from packy_agent.utils.installation import dump_version, remove_upgrade_in_progress_lock


logger = logging.getLogger(__name__)


def entry():
    settings.set_runtime('component', WATCHDOG_COMPONENT)

    configure_logging_basic(logging.WARNING)
    parser = get_base_argument_parser(__loader__.name)
    args = parser.parse_args()

    settings.set_commandline_arguments(vars(args))
    configure_logging_basic(args.log_level)
    init_sentry_client()
    configure_logging(args.log_level)

    remove_upgrade_in_progress_lock()
    dump_version(settings.get_watchdog_version_dump_filename(),
                 settings.get_watchdog_version_dump_variable_name())
    settings.update_started_at_ts()

    service = WatchdogService()
    signal.signal(signal.SIGTERM, lambda signum, frame: service.graceful_stop())
    service.run()


if __name__ == '__main__':
    sys.exit(entry())



import sys


from packy_agent.configuration.settings import settings
from packy_agent.utils.cli import get_base_argument_parser
from packy_agent.utils.logging import configure_logging, configure_logging_basic
from packy_agent.domain_logic.managers.network import network_manager


DEFAULT_LOG_LEVEL = 'ERROR'


def set_dhcp(interface, no_reboot, no_backup):
    kwargs = {'reboot_delay_seconds': None} if no_reboot else {}
    network_manager.set_dhcp(interface, no_backup=no_backup, **kwargs)


def entry():
    configure_logging_basic(DEFAULT_LOG_LEVEL)
    parser = get_base_argument_parser(__loader__.name, default_log_level=DEFAULT_LOG_LEVEL)
    parser.add_argument('-i', '--interface')
    parser.add_argument('--no-reboot', action='store_true')
    parser.add_argument('--no-backup', action='store_true')
    args = parser.parse_args()
    settings.set_commandline_arguments(vars(args))
    configure_logging(DEFAULT_LOG_LEVEL)

    return set_dhcp(args.interface, args.no_reboot, args.no_backup)


if __name__ == '__main__':
    sys.exit(entry())



import os

from packy_agent.configuration.settings import settings


def get_available_disk_space(path=None):
    path = path or settings.get_database_filename()
    statvfs = os.statvfs(path)
    return statvfs.f_frsize * statvfs.f_bavail



import os
import distro


LINUX_MINT = 'linuxmint'
UBUNTU = 'ubuntu'
RASPBIAN = 'raspbian'
ALPINE = 'alpine'

PRETTY_NAMES = {
    LINUX_MINT: 'Linux Mint',
    UBUNTU: 'Ubuntu',
    RASPBIAN: 'Raspbian',
    ALPINE: 'Alpine',
}

LINUX_MINT_18_1 = (LINUX_MINT, '18.1')
UBUNTU_16_04 = (UBUNTU, '16.04')
UBUNTU_18_04 = (UBUNTU, '18.04')
RASPBIAN_9 = (RASPBIAN, '9')


def get_os_id_and_version():
    distro_id = distro.id()
    version_parts = distro.version_parts(best=True)
    if distro_id == RASPBIAN:
        version = version_parts[0]
    else:
        version = '.'.join(version_parts[:2])

    return distro_id, version


def is_inside_docker_container():
    return os.path.isfile('/.dockerenv')


def get_platform():
    platform = distro.name(pretty=True) or 'unidentified'
    if is_inside_docker_container():
        platform = f'Docker/{platform}'

    return platform



from abc import ABC

from gevent.queue import Queue, Empty


from packy_agent.configuration.settings import settings
from packy_agent.worker.loops.base.misc import StartableStoppableLoop
from packy_agent.exceptions import ImproperlyConfiguredError


class ConsumerLoop(StartableStoppableLoop, ABC):

    def __init__(self, inbound_queue=None, callable_=None, timeout=None):
        super().__init__()
        self.inbound_queue = inbound_queue or Queue()
        self.callable = callable_
        self.timeout = timeout or settings.get_worker_consumer_loop_timeout()

    def process_item(self, item):
        if self.callable:
            return self.callable(item)
        else:
            raise ImproperlyConfiguredError("Provide callable or override 'process_item' method")

    def iteration(self):
        try:
            item = self.inbound_queue.get(timeout=self.timeout)
        except Empty:
            return

        self.process_item(item)



import socket
import time
import logging
import random

import cachetools
from ryu.lib.packet.ipv4 import ipv4
from ryu.lib.packet.udp import udp
from ryu.lib.packet.icmp import ICMP_TIME_EXCEEDED, ICMP_DEST_UNREACH, ICMP_PORT_UNREACH_CODE
from ryu.lib.packet.packet import Packet

from packy_agent.domain_logic.modules.traceroute.base import Traceroute
from packy_agent.domain_logic.modules.traceroute.models import SentProbe
from packy_agent.domain_logic.modules.common.socket import (
    TIME_SIZE_BYTES, encode_timestamp, decode_timestamp)
from packy_agent.domain_logic.modules.common.constants import (
    IP_PACKET_HEADER_SIZE_BYTES, UDP_PACKET_HEADER_SIZE_BYTES)
from packy_agent.utils.misc import generate_random_string, safe_hexlify
from packy_agent.exceptions import OutOfPortsError


PACKET_HEADERS_SIZE_BYTES = IP_PACKET_HEADER_SIZE_BYTES + UDP_PACKET_HEADER_SIZE_BYTES
UDP_PORT_LOW = 33434
UDP_PORT_HIGH = 35357 - 1  # `OpenStack Identity (Keystone) administration` minus one
# These ports are unofficially assigned to Jenkins, Infestation: Survivor Stories, Factorio
UDP_PORT_GAPS = (33848, 34000, 34197)  # these ports are uno

TTL_CACHE_SIZE = UDP_PORT_HIGH - UDP_PORT_LOW - len(UDP_PORT_GAPS) + 1
used_udp_ports = cachetools.TTLCache(TTL_CACHE_SIZE, ttl=60)
used_udp_ports[UDP_PORT_LOW] = None


logger = logging.getLogger(__name__)


def reserve_udp_port():
    while True:
        used_udp_ports_len = len(used_udp_ports)
        if used_udp_ports_len >= TTL_CACHE_SIZE:
            raise OutOfPortsError()

        # It appears that using higher port numbers produce strange results therefore using this
        # formula use lower port number if possible
        high = min(UDP_PORT_LOW + (used_udp_ports_len * 2) + 1, UDP_PORT_HIGH + 1)
        port = random.randrange(UDP_PORT_LOW, high, 1)
        if port in used_udp_ports or port in UDP_PORT_GAPS:
            time.sleep(0.001)
            continue

        used_udp_ports[port] = None
        logger.debug('RESERVED UDP port: %s', port)
        return port


def free_udp_port(port):
    logger.debug('FREED UDP port: %s', port)
    used_udp_ports.pop(port, None)


def send_udp_probe(raw_socket, destination_ip_address, port, ttl, packet_size_bytes=60,
                   payload=None):
    logger.debug('Sending probe UDP packet to %s:%s (TTL:%s)...', destination_ip_address, port, ttl)
    raw_socket.setsockopt(socket.SOL_IP, socket.IP_TTL, ttl)

    sent_time = time.time()
    time_part = encode_timestamp(sent_time)
    if payload is None:
        if packet_size_bytes < PACKET_HEADERS_SIZE_BYTES:
            raise ValueError('payload_size_bytes must be greater or equal to {}'.format(
                PACKET_HEADERS_SIZE_BYTES))

        payload_size_bytes = packet_size_bytes - PACKET_HEADERS_SIZE_BYTES

        if payload_size_bytes < TIME_SIZE_BYTES:
            payload = generate_random_string(payload_size_bytes).encode('ascii')
        else:
            payload = time_part + generate_random_string(
                payload_size_bytes - TIME_SIZE_BYTES).encode('ascii')
    else:
        payload = time_part + payload

    raw_socket.sendto(payload, (destination_ip_address, port))
    logger.debug('Sent UDP packet to {}:{} (TTL:{}), payload: {}'.format(
        destination_ip_address, port, ttl, safe_hexlify(payload)))
    return sent_time, payload


class TracerouteUDP(Traceroute):

    range_min = PACKET_HEADERS_SIZE_BYTES

    def send_probe(self, probe_number, destination_ip_address, ttl, packet_size_bytes=60, **kwargs):
        port = reserve_udp_port()
        sent_time, sent_payload = send_udp_probe(
            self.get_sending_socket(), destination_ip_address=destination_ip_address,
            port=port, ttl=ttl, packet_size_bytes=packet_size_bytes)
        return SentProbe(ttl, probe_number, sent_time, sent_payload, port)

    @staticmethod
    def is_probe_reply(icmp_packet):
        return icmp_packet.type in (ICMP_TIME_EXCEEDED, ICMP_DEST_UNREACH)

    @staticmethod
    def retrieve_probe_packet(icmp_packet):
        inner_packet = Packet(icmp_packet.data.data, parse_cls=ipv4)
        protocol_1 = inner_packet.protocols[1]
        if isinstance(protocol_1, udp):
            udp_packet = protocol_1
            logger.debug('Processing UDP packet port: %s', udp_packet.dst_port)
            return udp_packet
        else:
            logger.debug('ICMP packet does not contain a UDP packet')
            return

    def process_method_specific_probe_reply(self, ip_packet, icmp_packet, probe_packet):
        probe_port = probe_packet.dst_port
        if probe_port not in self.sent_probes:
            logger.debug('Discarded ICMP packet (reason: unexpected port %s) '
                         'from: %s, IP checksum: %s, type: %s, code: %s', probe_port,
                         ip_packet.src, ip_packet.csum, icmp_packet.type, icmp_packet.code)
            return False

        return True

    @staticmethod
    def retrieve_payload(icmp_packet, probe_packet):
        inner_packet = Packet(icmp_packet.data.data, parse_cls=ipv4)
        if len(inner_packet.protocols) > 2:
            # TODO(dmu) HIGH: Is it possible to retrieve payload from udp_packet?
            return inner_packet.protocols[2]

    def get_sent_probe(self, probe_packet):
        return self.sent_probes.get(probe_packet.dst_port)

    @staticmethod
    def is_last_hop(icmp_packet):
        if icmp_packet.type == ICMP_TIME_EXCEEDED:
            return False
        elif icmp_packet.type == ICMP_DEST_UNREACH:
            # weird, but it means that destination is reached if we get a destination unreachable
            # ICMP packet with code ICMP_PORT_UNREACH_CODE
            return True
        else:
            raise ValueError(
                f'Unsupported ICMP packet type: {icmp_packet.type} (should never get here)')

    @staticmethod
    def retrieve_sent_time_from_payload(payload):
        if payload and len(payload) >= TIME_SIZE_BYTES:
            return decode_timestamp(payload[:TIME_SIZE_BYTES])

    @staticmethod
    def get_hop_ip_address(ip_packet, icmp_packet):
        if icmp_packet.type == ICMP_DEST_UNREACH and icmp_packet.code != ICMP_PORT_UNREACH_CODE:
            return
        else:
            return ip_packet.src

    def free_resources(self, probe_result):
        free_udp_port(probe_result.probe.key)

    def open_sending_socket(self):
        logger.debug('Opening sending socket...')
        try:
            return socket.socket(socket.AF_INET, socket.SOCK_DGRAM, self.udp_protocol)
        except socket.error as ex:
            if ex.errno == 1:  # Operation not permitted
                # TODO(dmu) HIGH: Fix to receive ICMP packet without being root
                raise socket.error(
                    ex.strerror + '  - process should be running as root.')

            raise



from collections import namedtuple

ProbeResult = namedtuple('ProbeResult', ('hop_number', 'reply_hop_number', 'probe_number',
                                         'hop_ip_address', 'rtt_seconds', 'is_last_hop',
                                         'probe'))
SentProbe = namedtuple('SentProbe', ('ttl', 'probe_number', 'sent_time', 'payload', 'key'))



import logging
import math
import select
import socket
from abc import ABCMeta, abstractmethod
import time
from collections import defaultdict

from packy_agent.domain_logic.modules.common.misc import retrieve_ip_and_icmp_packets
from packy_agent.domain_logic.modules.traceroute.constants import ICMP_METHOD, UDP_METHOD
from packy_agent.domain_logic.modules.common.constants import (
    IP_PACKET_HEADER_SIZE_BYTES, ICMP_PACKET_HEADER_SIZE_BYTES, UDP_PACKET_HEADER_SIZE_BYTES)
from packy_agent.domain_logic.modules.common.buffer import (
    buffer_icmp_response, pop_buffered_icmp_response_receive_time)
from packy_agent.domain_logic.modules.traceroute.models import ProbeResult
from packy_agent.exceptions import ImproperlyConfiguredError
from packy_agent.utils.misc import safe_hexlify


MIN_ICMP_METHOD_PACKET_SIZE = IP_PACKET_HEADER_SIZE_BYTES + ICMP_PACKET_HEADER_SIZE_BYTES
MIN_UDP_METHOD_PACKET_SIZE = IP_PACKET_HEADER_SIZE_BYTES + UDP_PACKET_HEADER_SIZE_BYTES
RANGE_MIN = {
    ICMP_METHOD: MIN_ICMP_METHOD_PACKET_SIZE,
    UDP_METHOD: MIN_UDP_METHOD_PACKET_SIZE,
}
MAX_PACKET_SIZE = 1500
MIN_PERIOD_DURATION_SECONDS = 0.001

logger = logging.getLogger(__name__)


def guess_reply_hop_number(ttl):
    if ttl <= 64:
        return 65 - ttl
    elif ttl <= 128:
        return 129 - ttl
    else:
        return 256 - ttl


def aggregate_results(trace_results):
    agg_results = []
    for probe_results in trace_results:
        aggregated = {'loss': probe_results.count(None)}
        filtered_probe_results = list(filter(None, probe_results))
        if filtered_probe_results:
            # TODO(dmu) HIGH: Host may be different for the different probes of the same TTL
            representative_probe = filtered_probe_results[0]
            aggregated['host'] = representative_probe.hop_ip_address
            aggregated['reply_hop_number'] = representative_probe.reply_hop_number
            rtts = tuple(x.rtt_seconds for x in filtered_probe_results)
            aggregated['last'] = rtts[-1]
            aggregated['best'] = min(rtts)
            aggregated['worst'] = max(rtts)
            average = sum(rtts) / len(rtts)
            aggregated['average'] = average

            variance = sum((x - average) ** 2 for x in rtts) / len(rtts)
            aggregated['stdev'] = math.sqrt(variance)

        agg_results.append(aggregated)

    return agg_results


def get_ttl_and_probe_number(max_hops, probe_count):
    for ttl in range(1, max_hops + 1):
        for probe_number in range(probe_count):
            yield (ttl, probe_number)


def does_payload_match(actual_payload, expected_payload):
    payload_len = len(expected_payload)
    if actual_payload:
        actual_payload_len = len(actual_payload)
        if actual_payload_len != payload_len:
            logger.warning('Payload length (%s bytes) differ from expected (sent) '
                           'length (%s bytes), but we are forgiving', actual_payload_len, payload_len)
            min_len = min(actual_payload_len, payload_len)
            actual_payload_prefix = actual_payload[:min_len]
            expected_payload_prefix = expected_payload[:min_len]
            if actual_payload_prefix != expected_payload_prefix:
                logger.debug('Payload prefix differs (actual: %s, expected: %s)',
                             safe_hexlify(actual_payload_prefix),
                             safe_hexlify(expected_payload_prefix))
                return False
        elif actual_payload != expected_payload:
            logger.debug('Payload differs (actual: %s, expected: %s)',
                         safe_hexlify(actual_payload), safe_hexlify(expected_payload))
            return False

    return True


class Traceroute(metaclass=ABCMeta):
    range_min = None

    def __init__(self):
        self.sending_socket = None
        self.receiving_socket = None
        # TODO(dmu) HIGH: Take care of memory leak here!
        self.sent_probes = {}

        self.udp_protocol = socket.getprotobyname('udp')
        self.icmp_protocol = socket.getprotobyname('icmp')

    @abstractmethod
    def send_probe(self, probe_number, destination_ip_address, ttl, packet_size_bytes, **kwargs):
        raise NotImplementedError('This method must be implemented in child class')

    @staticmethod
    @abstractmethod
    def is_probe_reply(icmp_packet):
        raise NotImplementedError('This method must be implemented in child class')

    @staticmethod
    @abstractmethod
    def retrieve_probe_packet(icmp_packet):
        raise NotImplementedError('This method must be implemented in child class')

    @abstractmethod
    def process_method_specific_probe_reply(self, ip_packet, icmp_packet, probe_packet):
        raise NotImplementedError('This method must be implemented in child class')

    @staticmethod
    @abstractmethod
    def retrieve_payload(icmp_packet, probe_packet):
        raise NotImplementedError('This method must be implemented in child class')

    @abstractmethod
    def get_sent_probe(self, probe_packet):
        raise NotImplementedError('This method must be implemented in child class')

    @staticmethod
    @abstractmethod
    def is_last_hop(icmp_packet):
        raise NotImplementedError('This method must be implemented in child class')

    @staticmethod
    @abstractmethod
    def retrieve_sent_time_from_payload(payload):
        raise NotImplementedError('This method must be implemented in child class')

    @staticmethod
    @abstractmethod
    def get_hop_ip_address(ip_packet, icmp_packet):
        raise NotImplementedError('This method must be implemented in child class')

    def free_resources(self, probe_result):
        pass

    def process_probe_reply(self, data, receive_time):
        logger.debug('PROCESSING probe reply received at %.6f: %s',
                     receive_time, safe_hexlify(data))

        result = retrieve_ip_and_icmp_packets(data)
        if result:
            ip_packet, icmp_packet = result
        else:
            return

        if not self.is_probe_reply(icmp_packet):
            logger.debug('Discarded ICMP packet (reason: not a probe reply) '
                         'from: %s, IP checksum: %s, type: %s, code: %s',
                         ip_packet.src, ip_packet.csum, icmp_packet.type, icmp_packet.code)
            return

        probe_packet = self.retrieve_probe_packet(icmp_packet)
        if not probe_packet:
            logger.debug(
                'Discarded ICMP packet (reason: does not contain proper probe packet) '
                'from: %s, IP checksum: %s, type: %s, code: %s',
                ip_packet.src, ip_packet.csum, icmp_packet.type, icmp_packet.code)
            return

        if not self.process_method_specific_probe_reply(ip_packet, icmp_packet, probe_packet):
            return

        actual_payload = self.retrieve_payload(icmp_packet, probe_packet)

        probe = self.get_sent_probe(probe_packet)
        assert probe
        expected_payload = probe.payload
        if not does_payload_match(actual_payload, expected_payload):
            logger.debug('Discarded ICMP packet (reason: payload does not match) '
                         'from: %s, IP checksum: %s, type: %s, code: %s',
                         ip_packet.src, ip_packet.csum, icmp_packet.type, icmp_packet.code)
            return

        logger.debug(
            'Accepted ICMP packet from: %s, IP checksum: %s, type: %s, code: %s, payload: %s',
            ip_packet.src, ip_packet.csum, icmp_packet.type, icmp_packet.code,
            safe_hexlify(actual_payload))

        # TODO(dmu) HIGH: Do we really need to retrieve sent time from payload?
        sent_time_from_payload = self.retrieve_sent_time_from_payload(actual_payload)
        if sent_time_from_payload is None:
            actual_sent_time = probe.sent_time
        else:
            actual_sent_time = sent_time_from_payload

        rtt_seconds = receive_time - actual_sent_time
        if rtt_seconds < 0:
            logger.error('rtt_seconds is less than zero')

        return ProbeResult(
            hop_number=probe.ttl,
            reply_hop_number=guess_reply_hop_number(ip_packet.ttl),
            probe_number=probe.probe_number,
            hop_ip_address=self.get_hop_ip_address(ip_packet, icmp_packet),
            rtt_seconds=rtt_seconds,
            is_last_hop=self.is_last_hop(icmp_packet),
            probe=probe)

    def get_probe_result(self):
        logger.debug('Getting probe result...')
        receiving_socket = self.get_receiving_socket()

        ready_sockets = select.select((receiving_socket,), (), (), 0.001)
        # because by this time the entire packet is already in OS buffer
        receive_time = time.time()

        if ready_sockets[0]:
            logger.debug('RECEIVING probe reply from socket...')
            packet_data = receiving_socket.recv(2048)
            logger.debug('RECEIVED data from socket at %.6f: %s',
                         receive_time, safe_hexlify(packet_data))
            probe_result = self.process_probe_reply(packet_data, receive_time)
            if probe_result:
                logger.debug('RECEIVED probe reply from socket: %s', probe_result)
                buffered_receive_time = pop_buffered_icmp_response_receive_time(packet_data)
                if buffered_receive_time is not None:
                    time_delta = receive_time - buffered_receive_time
                    logger.debug('Probe reply has been received %.6f seconds earlier', time_delta)
                    probe_result = ProbeResult(
                        hop_number=probe_result.hop_number,
                        reply_hop_number=probe_result.reply_hop_number,
                        probe_number=probe_result.probe_number,
                        hop_ip_address=probe_result.hop_ip_address,
                        rtt_seconds=probe_result.rtt_seconds - time_delta,
                        is_last_hop=probe_result.is_last_hop,
                        probe=probe_result.probe)

                self.free_resources(probe_result)
                return probe_result
            else:
                buffer_icmp_response(packet_data, receive_time)
        else:
            logger.debug('NOTHING on socket')

    def do(self, host, timeout=2, probe_count=1, packet_size_bytes=60, max_hops=100,
           max_parallelism=10):
        logger.debug('Starting %s: host=%s, timeout=%s, probe_count=%s, packet_size_bytes=%s, '
                     'max_hops=%s, max_parallelism=%s', self.__class__.__name__, host, timeout,
                     probe_count, packet_size_bytes, max_hops, max_parallelism)
        self.validate_arguments(packet_size_bytes, max_parallelism)
        destination_ip_address = self.get_destination_ip_address(host)
        if not destination_ip_address:
            return

        last_hop_number = max_hops
        results = defaultdict(dict)
        timeout_expiration = None
        active_probes = set()
        is_done = False
        should_send = True
        ttl_and_probe_number_generator = get_ttl_and_probe_number(max_hops, probe_count)
        while not is_done:
            start = time.time()
            if should_send:
                if len(active_probes) < max_parallelism:
                    try:
                        ttl, probe_number = next(ttl_and_probe_number_generator)
                    except StopIteration:
                        logger.debug('Sent out all probes')
                        should_send = False
                    else:
                        should_send = ttl <= last_hop_number
                        if should_send:
                            probe = self.send_probe(probe_number, destination_ip_address, ttl,
                                                    packet_size_bytes)
                            self.sent_probes[probe.key] = probe
                            active_probes.add(probe)
                        else:
                            logger.debug('Last hop has been reached: %s', last_hop_number)
                else:
                    logger.debug('No more room for parallelism')
            else:
                if timeout_expiration is None:
                    timeout_expiration = time.time() + timeout
                logger.debug('No more probes to send')

            while self.sent_probes:
                probe_result = self.get_probe_result()
                if probe_result:
                    logger.debug('Got probe result: %s', probe_result)
                    del self.sent_probes[probe_result.probe.key]
                else:
                    logger.debug('No probe result yet, moving on')
                    break

                # We got probe result but if does not have hop IP address then it was not sent by the hop
                results[probe_result.hop_number][probe_result.probe_number] = (
                    probe_result if probe_result.hop_ip_address else None)

                if (probe_result.is_last_hop or
                        probe_result.hop_ip_address == destination_ip_address):
                    if probe_result.hop_number < last_hop_number:
                        last_hop_number = probe_result.hop_number
                        logger.debug('CHANGED last hop number to: %s', last_hop_number)

                for ttl_local in range(last_hop_number, 0, -1):
                    results_len = len(results[ttl_local])
                    if results_len < probe_count:
                        logger.debug(
                            'Not all results have arrived yet (only %s replies for TTL: %s)',
                            results_len, ttl_local)
                        break
                else:
                    logger.debug('All results have arrived')
                    is_done = True
                    break

            for active_probe in list(active_probes):
                if time.time() >= active_probe.sent_time + timeout:
                    logger.debug('Probe %s timed out', active_probe)
                    active_probes.discard(active_probe)

            if not should_send and not self.sent_probes:
                logger.debug('No more probes to send or replies to receive')
                break

            if timeout_expiration is not None and time.time() >= timeout_expiration:
                logger.debug('Timed out')
                break

            duration = time.time() - start
            if duration < MIN_PERIOD_DURATION_SECONDS:
                sleep_duration = MIN_PERIOD_DURATION_SECONDS - duration
                logger.debug('Sleeping for %.6g seconds to stay IO-bound', sleep_duration)
                time.sleep(sleep_duration)

        self.sent_probes.clear()

        # Cut the tail (for unreachable host case)
        new_last_hop_number = last_hop_number
        for ttl in range(last_hop_number, 0, -1):
            new_last_hop_number = ttl
            result = results.get(ttl)
            if result and any(result.values()):
                new_last_hop_number = min(new_last_hop_number + 1, last_hop_number)
                break

        last_hop_number = new_last_hop_number

        return aggregate_results(
            [[results[ttl].get(probe_number) for probe_number in range(probe_count)]
                for ttl in range(1, last_hop_number + 1)])

    @abstractmethod
    def open_sending_socket(self):
        raise NotImplementedError('This method must be implemented in child class')

    def open_receiving_socket(self):
        logger.debug('Opening receiving socket...')
        try:
            return socket.socket(socket.AF_INET, socket.SOCK_RAW, self.icmp_protocol)
        except socket.error as ex:
            if ex.errno == 1:  # Operation not permitted
                # TODO(dmu) HIGH: Fix to receive ICMP packet without being root
                raise socket.error(
                    ex.strerror + '  - process should be running as root.')

            raise

    @staticmethod
    def get_destination_ip_address(host):
        try:
            return socket.gethostbyname(host)
        except socket.gaierror:
            logger.warning('Could not resolve {} to IP address'.format(host))
            return

    def validate_arguments(self, packet_size_bytes, max_parallelism):
        if self.range_min is None:
            raise ImproperlyConfiguredError('`range_min` attribute must be set in child class')

        if max_parallelism < 1:
            raise ValueError('max_parallelism must be greater or equal to 1')

        range_min = self.range_min
        if not (self.range_min <= packet_size_bytes <= MAX_PACKET_SIZE):
            raise ValueError(
                'packet_size_bytes must be in range from {} to 1500 (inclusive)'.format(range_min))

    def get_sending_socket(self):
        self.get_receiving_socket()  # will open receiving socket and make sure no reply is missing
        if not self.sending_socket:
            self.sending_socket = self.open_sending_socket()

        return self.sending_socket

    def get_receiving_socket(self):
        if not self.receiving_socket:
            self.receiving_socket = self.open_receiving_socket()

        return self.receiving_socket

    def close_sending_socket(self):
        if self.sending_socket:
            try:
                self.sending_socket.close()
            except Exception as ex:
                logger.warning('Error while closing sending socket: %r', ex)

            self.sending_socket = None

    def close_receiving_socket(self):
        if self.receiving_socket:
            try:
                self.receiving_socket.close()
            except Exception as ex:
                logger.warning('Error while closing receiving socket: %r', ex)

            self.receiving_socket = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close_sending_socket()
        self.close_receiving_socket()


def traceroute(host, timeout=3, probe_count=1, packet_size_bytes=60, max_hops=100,
               method=UDP_METHOD, max_parallelism=10):

    if method not in (UDP_METHOD, ICMP_METHOD):
        raise ValueError('Unknown traceroute method: {}'.format(method))

    if max_parallelism < 1:
        raise ValueError('max_parallelism must be greater or equal to 1')

    range_min = RANGE_MIN[method]
    if not (range_min <= packet_size_bytes <= MAX_PACKET_SIZE):
        raise ValueError(
            'packet_size_bytes must be in range from {} to 1500 (inclusive)'.format(range_min))

    if method == UDP_METHOD:
        from packy_agent.domain_logic.modules.traceroute.methods.udp import TracerouteUDP
        traceroute = TracerouteUDP()
    elif method == ICMP_METHOD:
        from packy_agent.domain_logic.modules.traceroute.methods.icmp import TracerouteICMP
        traceroute = TracerouteICMP()
    else:
        assert False

    with traceroute:
        return traceroute.do(host, timeout=timeout, probe_count=probe_count,
                             packet_size_bytes=packet_size_bytes, max_hops=max_hops,
                             max_parallelism=max_parallelism)



class ServiceBase:

    def __init__(self, name):
        self.name = name

    def start(self):
        raise NotImplementedError('Must be implemented in child class')

    def stop(self):
        raise NotImplementedError('Must be implemented in child class')

    def restart(self):
        raise NotImplementedError('Must be implemented in child class')

    def is_running(self):
        raise NotImplementedError('Must be implemented in child class')



import pytest

from packy_agent.configuration.sources.base import MappingBackedSource


def test_mapping_backed_source():
    source = MappingBackedSource({'value1': 10, 'value2': 15})
    assert set(source) == {'value1', 'value2'}
    assert source['value1'] == 10
    assert source['value2'] == 15
    assert 'value1' in source
    assert source.get('value1') == 10
    assert source.get('value3', 4) == 4
    assert sorted(source.keys()) == ['value1', 'value2']
    assert sorted(source.values()) == [10, 15]
    assert sorted(source.items()) == [('value1', 10), ('value2', 15)]

    with pytest.raises(TypeError) as ex:
        source['value1'] = 0

    assert "'MappingBackedSource' object does not support item assignment" in str(ex.value)

    with pytest.raises(TypeError) as ex:
        del source['value1']
    assert "'MappingBackedSource' object does not support item deletion" in str(ex.value)

    with pytest.raises(AttributeError) as ex:
        source.pop('value1')
    assert "'MappingBackedSource' object has no attribute 'pop'" in str(ex.value)



from flask_wtf import FlaskForm
from wtforms import StringField, BooleanField, HiddenField
from wtforms import validators


class CommaSeparatedIPAddresses(validators.IPAddress):
    def __call__(self, form, field):
        value = field.data
        valid = False
        if value:
            for ip_address in value.split(','):
                ip_address = ip_address.strip()
                valid = ((self.ipv4 and self.check_ipv4(ip_address)) or
                         (self.ipv6 and self.check_ipv6(ip_address)))

        if not valid:
            message = self.message
            if message is None:
                message = field.gettext('Invalid IP address.')
            raise validators.ValidationError(message)


class NetworkForm(FlaskForm):

    dhcp = BooleanField('Use DHCP')
    ip_address = StringField('IP Address', validators=[validators.Optional(),
                                                       validators.IPAddress()])
    subnet_mask = StringField('Subnet Mask', validators=[validators.Optional(),
                                                         validators.IPAddress()])
    default_gateway = StringField('Default Gateway', validators=[validators.Optional(),
                                                                 validators.IPAddress()])
    name_servers = StringField('Name Servers', validators=[validators.Optional(),
                                                           CommaSeparatedIPAddresses()])

    network_interface = HiddenField()

    def validate(self):
        rv = super(NetworkForm, self).validate()
        if not rv:
            return False

        is_valid = True
        if not self.dhcp.data:
            for field in (self.ip_address, self.subnet_mask, self.default_gateway):
                if not field.raw_data or not field.raw_data[0]:
                    field.errors.append(field.gettext('This field is required.'))
                    is_valid = False

        return is_valid



import time
import json
from unittest.mock import MagicMock

import httpretty
import pytest

from packy_agent.watchdog.service import WatchdogService
from packy_agent.configuration.settings import settings
from packy_agent.tests.fixtures.configuration import override_settings


loops = {
    'ping_module': {
        'counter': 1,
        'is_greenlet_dead': False,
        'is_running': True,
        'loop_type': 'scheduled_producer',
        'produced_counter': 1,
        'schedule': '*/1 * * * * *'
    }
}


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_smoke():
    with httpretty.enabled():
        url = 'http://{}:{}/status/'.format(settings.get_worker_http_bind_address(),
                                            settings.get_worker_http_port())
        httpretty.register_uri(httpretty.GET, url, status=200, body=json.dumps({'loops': loops}))
        httpretty.register_uri(httpretty.GET, settings.get_console_base_url(),
                               status=200)  # emulate console running

        service = WatchdogService()

        service.reboot = MagicMock()
        service.restart_worker = MagicMock()
        service.start_worker = MagicMock()

        service.resolve_online_status = MagicMock(return_value=None)
        service.resolve_worker_supervisor_program_status = MagicMock(return_value=None)
        service.resolve_worker_process_status = MagicMock(return_value=None)
        service.resolve_loops_activity = MagicMock(return_value=None)
        service.report_error = MagicMock()
        service.report_warning = MagicMock()

        service.loop_iteration()

        service.start_worker.assert_not_called()
        service.restart_worker.assert_not_called()
        service.reboot.assert_not_called()

        service.report_warning.assert_not_called()
        service.report_error.assert_not_called()


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_smoke_2_iterations():
    with httpretty.enabled():
        url = 'http://{}:{}/status/'.format(settings.get_worker_http_bind_address(),
                                            settings.get_worker_http_port())
        httpretty.register_uri(httpretty.GET, url, status=200, body=json.dumps({'loops': loops}))
        httpretty.register_uri(httpretty.GET, settings.get_console_base_url(),
                               status=200)  # emulate console running

        service = WatchdogService()

        service.reboot = MagicMock()
        service.restart_worker = MagicMock()
        service.start_worker = MagicMock()

        service.resolve_online_status = MagicMock(return_value=None)
        service.resolve_worker_supervisor_program_status = MagicMock(return_value=None)
        service.resolve_worker_process_status = MagicMock(return_value=None)
        service.resolve_loops_activity = MagicMock(return_value=None)
        service.report_error = MagicMock()
        service.report_warning = MagicMock()

        service.loop_iteration()
        time.sleep(0.2)
        service.loop_iteration()

        service.start_worker.assert_not_called()
        service.restart_worker.assert_not_called()
        service.reboot.assert_not_called()

        service.report_warning.assert_not_called()
        service.report_error.assert_not_called()


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_collection_stale():
    with override_settings({'watchdog': {'relax_period_seconds': 0}}):
        with httpretty.enabled():
            url = 'http://{}:{}/status/'.format(settings.get_worker_http_bind_address(),
                                                settings.get_worker_http_port())
            httpretty.register_uri(httpretty.GET, url, status=200,
                                   body=json.dumps({'loops': loops, 'collected_counter': 1}))
            httpretty.register_uri(httpretty.GET, settings.get_console_base_url(),
                                   status=200)  # emulate console running

            service = WatchdogService()

            service.reboot = MagicMock()
            service.restart_worker = MagicMock()
            service.start_worker = MagicMock()

            service.resolve_online_status = MagicMock(return_value=None)
            service.resolve_worker_supervisor_program_status = MagicMock(return_value=None)
            service.resolve_worker_process_status = MagicMock(return_value=None)
            service.report_error = MagicMock()
            service.report_warning = MagicMock()

            service.loop_iteration()
            time.sleep(1.5)
            service.loop_iteration()

            service.start_worker.assert_not_called()
            service.restart_worker.assert_not_called()
            service.reboot.assert_not_called()

            service.report_error.assert_not_called()
            service.report_warning.assert_called()
            assert service.report_warning.call_args_list
            assert any(args[0][0].startswith('Measurement collection is stale on ') for args
                       in service.report_warning.call_args_list)


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_submission_stale():
    with override_settings({'watchdog': {'relax_period_seconds': 0}}):
        with httpretty.enabled():
            url = 'http://{}:{}/status/'.format(settings.get_worker_http_bind_address(),
                                                settings.get_worker_http_port())
            httpretty.register_uri(httpretty.GET, url, status=200,
                                   body=json.dumps({'loops': loops, 'collected_counter': 1,
                                                    'submitted_counter': 1}))
            httpretty.register_uri(httpretty.GET, settings.get_console_base_url(),
                                   status=200)  # emulate console running

            service = WatchdogService()

            service.reboot = MagicMock()
            service.restart_worker = MagicMock()
            service.start_worker = MagicMock()

            service.resolve_online_status = MagicMock(return_value=None)
            service.resolve_worker_supervisor_program_status = MagicMock(return_value=None)
            service.resolve_worker_process_status = MagicMock(return_value=None)
            service.report_error = MagicMock()
            service.report_warning = MagicMock()

            service.loop_iteration()
            time.sleep(1.5)
            httpretty.register_uri(httpretty.GET, url, status=200,
                                   body=json.dumps({'loops': loops, 'collected_counter': 2,
                                                    'submitted_counter': 1}))

            service.loop_iteration()

            service.start_worker.assert_not_called()
            service.restart_worker.assert_not_called()
            service.reboot.assert_not_called()

            service.report_error.assert_not_called()
            service.report_warning.assert_called()
            assert service.report_warning.call_args_list
            assert any(args[0][0].startswith('Measurement submission is stale on ') for args
                       in service.report_warning.call_args_list)


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_purge_stale():
    with override_settings({'watchdog': {'relax_period_seconds': 0},
                            'worker': {'purge_period_seconds': 1}}):
        with httpretty.enabled():
            url = 'http://{}:{}/status/'.format(settings.get_worker_http_bind_address(),
                                                settings.get_worker_http_port())
            httpretty.register_uri(httpretty.GET, url, status=200,
                                   body=json.dumps({'loops': loops, 'collected_counter': 1,
                                                    'submitted_counter': 1, 'purged_records': 10}))
            httpretty.register_uri(httpretty.GET, settings.get_console_base_url(),
                                   status=200)  # emulate console running

            service = WatchdogService()

            service.reboot = MagicMock()
            service.restart_worker = MagicMock()
            service.start_worker = MagicMock()

            service.resolve_online_status = MagicMock(return_value=None)
            service.resolve_worker_supervisor_program_status = MagicMock(return_value=None)
            service.resolve_worker_process_status = MagicMock(return_value=None)
            service.report_error = MagicMock()
            service.report_warning = MagicMock()

            service.loop_iteration()
            time.sleep(1.5)
            httpretty.register_uri(httpretty.GET, url, status=200,
                                   body=json.dumps({'loops': loops, 'collected_counter': 2,
                                                    'submitted_counter': 2, 'purged_records': 10}))

            service.loop_iteration()

            service.start_worker.assert_not_called()
            service.restart_worker.assert_not_called()
            service.reboot.assert_not_called()

            service.report_error.assert_not_called()
            service.report_warning.assert_called()
            assert service.report_warning.call_args_list
            assert any(args[0][0].startswith('Measurement purging is stale on ') for args
                       in service.report_warning.call_args_list)



import json

import pytest
import httpretty
from sqlalchemy import create_engine

from packy_agent.configuration.settings import settings
from packy_agent.domain_logic.models.sqlalchemy.base import Session
from packy_agent.domain_logic.models.sqlalchemy.measurement import Measurement
from packy_agent.tests.worker.loops.main.base import run_main_loop


TASKS = [
    {
        'id': 1,
        'module_name': 'ping',
        'args': [],
        'kwargs': {
            'number_of_pings': 3,
            'host': '8.8.8.8',
            'interval_ms': 0,
            'packet_size': 64
        },
        'schedule': '*/2 * * * * *'
    },
]


def assert_stored_value(value):
    assert isinstance(value['ts'], int)
    assert value['target'] == '8.8.8.8'
    assert value['n_pings'] == 3
    assert value['packet_size'] == 64
    assert isinstance(value['values'], list)
    if value['values']:
        assert isinstance(value['values'][0], (float, int))


@pytest.mark.usefixtures('activated')
@pytest.mark.parametrize('http_status', [500, 501, 502, 503, 504])
def test_measurements_keep_collecting_if_server_returns_http5xx(http_status):
    with httpretty.enabled():
        url = settings.get_server_base_url() + 'api/v2/agent/module/ping/measurement/'
        httpretty.register_uri(httpretty.POST, url, status=http_status)
        run_main_loop(TASKS)

    engine = create_engine(settings.get_database_url())
    with Session(engine) as session:
        measurements = session.query(Measurement).order_by(Measurement.created_at_ts).all()

    assert measurements
    measurements_iter = iter(measurements)
    first_measurement = next(measurements_iter)

    assert first_measurement.id
    assert first_measurement.measurement_type == 1
    assert first_measurement.created_at_ts
    assert first_measurement.error_at_ts
    assert first_measurement.error_message
    assert first_measurement.error_message.startswith(f'HTTP{http_status}:')
    assert not first_measurement.submitted_at_ts

    for measurement in measurements_iter:
        assert measurement.id
        assert measurement.measurement_type == 1
        assert measurement.created_at_ts
        assert not measurement.error_at_ts
        assert not measurement.error_message
        assert not measurement.submitted_at_ts

        assert_stored_value(json.loads(measurement.value))



import logging

import gevent
from gevent.pywsgi import WSGIServer
from gevent.greenlet import Greenlet
from flask import Flask
from flask import jsonify

from packy_agent.constants import STATUS_LOOP
from packy_agent.worker.loops.base.misc import StartableStoppableLoop
from packy_agent.worker.loops.base.periodic import PeriodicLoop
from packy_agent.worker.loops.base.scheduled import ScheduledLoop


logger = logging.getLogger(__name__)


class StatusLoop(StartableStoppableLoop):

    formal_name = STATUS_LOOP

    def __init__(self, main_loop, address='127.0.0.1', port=5000, **mkwargs):
        mkwargs.pop('is_safe_iteration', None)
        mkwargs.pop('is_logged_iteration', None)
        super().__init__(is_safe_iteration=False, is_logged_iteration=False, **mkwargs)

        self.main_loop = main_loop
        self.address = address
        self.port = port

        self._http_server = None

    def get_app(self):
        app = Flask(__name__)

        @app.route('/status/')
        def status():
            main_loop = self.main_loop

            loops = {}
            for loop in main_loop.loops:
                loop_greenlet = loop.get_greenlet()
                is_greenlet_dead = loop_greenlet.dead if loop_greenlet else None

                loop_properties = {
                    'is_running': getattr(loop, 'is_running', None),
                    'is_greenlet_dead': is_greenlet_dead,
                    'loop_type': loop.loop_type,
                    'counter': loop.counter,
                }

                if isinstance(loop, PeriodicLoop):
                    loop_properties['period'] = loop.period

                if isinstance(loop, ScheduledLoop):
                    loop_properties['schedule'] = loop.schedule

                produced_counter = getattr(loop, 'produced_counter', None)
                if produced_counter is not None:
                    loop_properties['produced_counter'] = produced_counter

                loops[loop.formal_name] = loop_properties

            task_results_consumer = main_loop.task_results_consumer
            if task_results_consumer:
                collected_counter = task_results_consumer.collected_counter
            else:
                collected_counter = None

            task_results_submitter = main_loop.task_results_submitter
            if task_results_submitter:
                submitted_counter = task_results_submitter.submitted_counter
            else:
                submitted_counter = None

            task_results_purger = main_loop.task_results_purger
            if task_results_purger:
                purged_records = task_results_purger.purged_records
            else:
                purged_records = None

            payload = {
                'loops': loops,
                'collected_counter': collected_counter,
                'submitted_counter': submitted_counter,
                'purged_records': purged_records,
            }
            return jsonify(payload)

        return app

    def iteration(self):
        gevent.sleep(60)  # faking it

    def get_greenlet(self):
        return

    def start(self):
        logger.info('STARTED %s', self.description)
        self.is_running = False
        self.is_stopping = False

        if not self._http_server:
            self._http_server = WSGIServer((self.address, self.port), self.get_app())

        if not self._http_server.started:
            self._http_server.start()

    def join(self):
        try:
            self._http_server._stop_event.wait()
        finally:
            Greenlet.spawn(self.stop).join()

        logger.info('STOPPED %s', self.description)

    def stop(self):
        super().stop()
        if self._http_server:
            self._http_server.stop()



import logging
from abc import abstractmethod, ABC
import time

logger = logging.getLogger(__name__)


class Loop(ABC):

    formal_name = None
    loop_type = 'base'

    def __init__(self, is_safe_iteration=True, is_logged_iteration=True):
        self.is_safe_iteration = is_safe_iteration
        self.is_logged_iteration = is_logged_iteration

        self.counter = 0

    @abstractmethod
    def iteration(self):
        raise NotImplementedError('Method must be implemented in child class')

    def iteration_wrapper(self):
        description = self.description
        if self.is_logged_iteration:
            logger.debug(f'STARTED iteration: {description}')

        start_time = time.time()
        try:
            self.iteration()
        except Exception:
            elapsed = time.time() - start_time
            if self.is_safe_iteration:
                logger.exception(f'ERROR during iteration (in {elapsed:.6g}s): {description}')
            else:
                raise
        else:
            elapsed = time.time() - start_time
            if self.is_logged_iteration:
                logger.debug(f'FINISHED iteration (in {elapsed:.6g}s): {description}')
        finally:
            self.counter += 1

    def loop(self):
        logger.debug(f'STARTED {self.description}')
        while True:
            self.iteration_wrapper()

    @property
    def description(self):
        return self.__class__.__name__



import logging
import os.path
from abc import ABC, abstractmethod
from collections.abc import Mapping, MutableMapping

import yaml

from packy_agent.utils.misc import get_lazy_value, SENTINEL

logger = logging.getLogger(__name__)


class ReadOnlyEmptyDict(Mapping):
    def __getitem__(self, item):
        raise KeyError(item)

    def __iter__(self):
        return iter(())

    def __len__(self):
        return 0

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


READ_ONLY_EMPTY_DICT = ReadOnlyEmptyDict()


class ConfigurationSourceBase(ABC, Mapping):
    is_writeable = False

    def __init__(self, name=None):
        self._is_preparing = False
        self._name = name

    @property
    def is_preparing(self):
        return self._is_preparing

    @property
    @abstractmethod
    def is_ready(self):
        raise NotImplementedError('Must be implemented in child class')

    @property
    def name(self):
        return self._name or self.__class__.__name__


class MappingBackedSource(ConfigurationSourceBase):

    def __init__(self, mapping=None, name=None):
        super().__init__(name=name)

        self._mapping = None

        if mapping is not None:
            self.mapping = mapping

    def __getitem__(self, item):
        return self.mapping[item]

    def __iter__(self):
        return iter(self.mapping)

    def __len__(self):
        return len(self.mapping)

    def __contains__(self, item):
        return item in self.mapping

    def __eq__(self, other):
        return self.mapping == other

    def __ne__(self, other):
        return self.mapping != other

    def keys(self):
        return self.mapping.keys()

    def items(self):
        return self.mapping.items()

    def values(self):
        return self.mapping.values()

    def get(self, key, default=None):
        logger.debug(f'GET "{self.name}" -> {key}')
        value = self.mapping.get(key, default)
        repr_value = 'SENTINEL' if value is SENTINEL else repr(value)
        logger.debug(f'GOT "{self.name}" -> {key} = {repr_value}')
        return value

    def get_mapping(self):
        return READ_ONLY_EMPTY_DICT if self._mapping is None else self._mapping

    def set_mapping(self, value):
        if value is not None and not isinstance(value, Mapping):
            raise TypeError("'mapping' must be an instance of 'Mapping' or None, but it is '{}'".format(
                type(value).__name__))

        self._mapping = value

    @property
    def is_ready(self):
        return self._mapping is not None

    def ensure_ready(self):
        pass

    @property
    def mapping(self):
        return self.get_mapping()

    @mapping.setter
    def mapping(self, value):
        self.set_mapping(value)


class FileBackedSource(MappingBackedSource, ABC):

    def __init__(self, filename, name=None):

        self._filename = filename
        self._mtime = None
        super().__init__(name=name)

    @staticmethod
    @abstractmethod
    def deserialize(stream):
        raise NotImplementedError('Must be implemented in child class')

    @property
    def filename(self):
        self._filename = get_lazy_value(self._filename)
        return self._filename

    def ensure_ready(self):
        if self.is_ready or self.is_preparing:
            return

        self.get_mapping()

    def get_mapping(self):
        if self._is_preparing:
            return super().get_mapping()

        self._is_preparing = True

        filename = self.filename

        try:
            try:
                mtime = os.path.getmtime(filename)
                if self._mapping is None or self._mtime != mtime:
                    with open(filename) as f:
                        conf = self.deserialize(f)

                    if not isinstance(conf, Mapping):
                        raise TypeError(f'{self.filename} must be a mapping on a root level')

                    self.set_mapping(conf)
                    self._mtime = mtime
            except FileNotFoundError:
                pass
            except OSError:
                logger.exception('Error while reading settings file: %s', filename)

            return super().get_mapping()
        finally:
            self._is_preparing = False

    @property
    def name(self):
        return super().name + f' ({self.filename})'


class YamlFileBackedSource(FileBackedSource):

    @staticmethod
    def deserialize(stream):
        return yaml.load(stream)


class MutableMappingBackedSource(MappingBackedSource, MutableMapping):
    def __setitem__(self, key, value):
        self.mapping[key] = value

    def __delitem__(self, key):
        del self.mapping[key]

    def pop(self, key):
        return self.mapping.pop(key)

    def popitem(self):
        return self.mapping.popitem()

    def clear(self):
        self.mapping.clear()

    def update(self, *args, **kwargs):
        self.mapping.update(*args, **kwargs)

    def setdefault(self, *args, **kwargs):
        return self.mapping.setdefault(*args, **kwargs)



"""initial

Revision ID: 54f648ebdb64
Revises: 
Create Date: 2018-08-25 17:58:51.997145

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = '54f648ebdb64'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        'measurement',
        sa.Column('id', sa.BigInteger().with_variant(sa.Integer(), 'sqlite'),
                  autoincrement=True, nullable=False),
        sa.Column('measurement_type', sa.Integer(), nullable=False),
        sa.Column('created_at_ts', sa.Integer(), nullable=False),
        sa.Column('submitted_at_ts', sa.Integer(), nullable=True),
        sa.Column('error_at_ts', sa.Integer(), nullable=True),
        sa.Column('error_message', sa.UnicodeText, nullable=True),
        sa.Column('value', sa.BLOB, nullable=False),
        sa.PrimaryKeyConstraint('id')
    )


def downgrade():
    op.drop_table('measurement')



import logging
import socket
import time

from packy_agent.constants import PING_MODULE, PING_MODULE_LOOP
from packy_agent.domain_logic.modules.ping.base import ping_once
from packy_agent.domain_logic.models.schematics.ping import PingModuleMeasurement
from packy_agent.worker.loops.modules.base import (
    ScheduledProducerTaskLoop, get_target, get_packet_size, get_number_of_pings)

logger = logging.getLogger(__name__)


def ping(*args, **kwargs):

    target = get_target(kwargs, PING_MODULE)
    if not target:
        return

    packet_size = get_packet_size(kwargs, PING_MODULE)
    number_of_pings = get_number_of_pings(kwargs, PING_MODULE)

    interval_ms = kwargs.get('interval_ms')
    if interval_ms is None:
        interval_ms = 0
        logger.warning('Ping interval (interval_ms) keyword argument was not provided, '
                       'using %s as default (ping)', interval_ms)

    measurement = PingModuleMeasurement({
        'target': target,
        'packet_size': packet_size,
        'n_pings': number_of_pings,
        'values': [],
    })  # we place it here because it also saves time moment
    try:
        destination_ip_address = socket.gethostbyname(target)
    except socket.gaierror:
        logger.warning('Could not resolve {} to IP address'.format(target))
    else:
        for ping_no in range(1, number_of_pings + 1):
            logger.debug('STARTED %s ping', ping_no)
            start = time.time()
            try:
                rtt_seconds = ping_once(destination_ip_address, 10, packet_size_bytes=packet_size)
                logger.debug('FINISHED %s ping', ping_no)
                if rtt_seconds is None:
                    continue
            except Exception:
                logger.warning('ERROR during ping', exc_info=True)
                continue

            measurement.values_.append(round(rtt_seconds * 1000, 2))

            time_left = interval_ms / 1000. - (time.time() - start)
            if time_left > 0:
                logger.debug(f'WAITING {time_left:.03f} seconds for next ping')
                time.sleep(time_left)

    return measurement


class PingTaskLoop(ScheduledProducerTaskLoop):

    formal_name = PING_MODULE_LOOP

    def __init__(self, schedule, args=(), kwargs=None, outbound_queue=None):
        super().__init__(schedule=schedule, callable_=ping, args=args, kwargs=kwargs,
                         outbound_queue=outbound_queue)



import os
import os.path
import time

import packy_agent


def dump_version(filename, variable_name):
    with open(filename, 'w') as f:
        f.write('{}={}\n'.format(variable_name, packy_agent.__version__))


def is_upgrade_in_progress():
    from packy_agent.configuration.settings import settings
    upgrade_in_progress_lock_filename = settings.get_upgrade_in_progress_lock_filename()
    try:
        mtime = os.path.getmtime(upgrade_in_progress_lock_filename)
    except OSError:
        return False

    upgrade_max_duration_seconds = settings.get_upgrade_max_duration_seconds()
    # We need this check because we want to prohibit automatics restarts/reboots during upgrades,
    # but still want them in case upgrade fails and keeps the lockfile
    return time.time() - mtime < upgrade_max_duration_seconds


def remove_upgrade_in_progress_lock():
    from packy_agent.configuration.settings import settings
    try:
        os.remove(settings.get_upgrade_in_progress_lock_filename())
    except OSError:
        pass



import re
import shutil
import os.path
import logging
from io import StringIO

from jinja2 import Template
import netifaces

from packy_agent.utils.misc import atomic_write
from packy_agent.utils.pkg_resources import get_package_file_content
from packy_agent.utils.shell import run_command_async
from packy_agent.utils.platforms import (get_os_id_and_version, LINUX_MINT_18_1, UBUNTU_16_04,
                                         UBUNTU_18_04, RASPBIAN_9, is_inside_docker_container)
from packy_agent.configuration.settings import settings
from packy_agent.utils.network import get_interfaces
from packy_agent.domain_logic.managers.control import control_manager
from packy_agent.utils.network import get_actual_ip_address


STATIC_TEMPLATE_TYPE = 'static'
DHCP_TEMPLATE_TYPE = 'dhcp'

UBUNTU_DEBIAN_TEMPLATES = {
    STATIC_TEMPLATE_TYPE: 'configuration/templates/network/ubuntu-debian/static',
    DHCP_TEMPLATE_TYPE: 'configuration/templates/network/ubuntu-debian/dhcp'
}

OPERATING_SYSTEM_TEMPLATES = {
    LINUX_MINT_18_1: UBUNTU_DEBIAN_TEMPLATES,
    UBUNTU_16_04: UBUNTU_DEBIAN_TEMPLATES,
    UBUNTU_18_04: UBUNTU_DEBIAN_TEMPLATES,
    RASPBIAN_9: UBUNTU_DEBIAN_TEMPLATES,
}

STANZA_RE = re.compile(r'^\s*(?:iface|mapping|auto|allow-\S+|source|source-directory)')
IFACE_STANZA_RE = re.compile(r'^\s*iface\s+(?P<iface>[^\s:]+)(?::\S+)?\s+inet\s+(?P<method>\S+)')
AUTO_STANZA_RE = re.compile(r'^\s*(auto|allow-auto)\s+(?P<iface>\S+)')


BACKUP_FILENAME_TEMPLATE = '{}.packy.backup'

ATTRS_RE = {
    'ip_address': re.compile('\s*address\s+(\S+)'),
    'subnet_mask': re.compile('\s*netmask\s+(\S+)'),
    'default_gateway': re.compile('\s*gateway\s+(\S+)'),
    'name_servers': re.compile('\s*dns-nameservers\s+(.*)'),
}
ANY_SPACE_RE = re.compile('\s+')

logger = logging.getLogger(__name__)


def update_network(interface, delay_seconds=0):
    run_command_async(f'ip addr flush {interface}; systemctl restart networking.service',
                      delay_seconds=delay_seconds)


def get_backup_filename(original_filename):
    return BACKUP_FILENAME_TEMPLATE.format(original_filename)


def get_stanzas():
    configuration_filename = settings.get_interfaces_configuration_filename()

    stanzas = []

    lines = []
    stanzas.append(lines)
    with open(configuration_filename) as f:
        for line in f:
            match = STANZA_RE.match(line)
            if match:
                lines = []
                stanzas.append(lines)

            lines.append(line)

    return stanzas


def get_stanza_by_interface(interface, stanzas=None):
    stanzas = get_stanzas() if stanzas is None else stanzas
    for lines in stanzas:
        if lines:
            first_line = lines[0]
            match = IFACE_STANZA_RE.match(first_line)
            if match and match.group('iface') == interface:
                return lines


def get_stanza_attributes(stanza):
    attrs = {}
    for line in stanza:
        for attr_name, regex in ATTRS_RE.items():
            match = regex.match(line)
            if match:
                attrs[attr_name] = match.group(1)

    return attrs


def get_replacing_content(interface, is_dhcp, context=None):
    os_id_and_version = get_os_id_and_version()
    template_paths = OPERATING_SYSTEM_TEMPLATES.get(os_id_and_version)
    if not template_paths:
        raise NotImplementedError(f'Operating system with with this ID is not supported: '
                                  f'{os_id_and_version[0]}')

    template_path = template_paths[DHCP_TEMPLATE_TYPE if is_dhcp else STATIC_TEMPLATE_TYPE]
    template = Template(get_package_file_content('packy_agent', template_path).decode('utf-8'))

    context = context or {}
    context['interface'] = interface
    return '\n\n' + template.render(**context) + '\n\n'


def get_interfaces_configuration_file_stream(interface, replacing_content):
    stream = StringIO()

    for lines in get_stanzas():
        if lines:
            first_line = lines[0]
            match = AUTO_STANZA_RE.match(first_line)
            if match and match.group('iface') == interface:
                continue

            match = IFACE_STANZA_RE.match(first_line)
            if match and match.group('iface') == interface:
                stream.write(replacing_content)
                replacing_content = ''
                continue

        for line in lines:
            stream.writelines(line)

    stream.write(replacing_content)
    stream.seek(0)

    return stream


class NetworkManager:

    @staticmethod
    def backup_configuration():
        original_filename = settings.get_interfaces_configuration_filename()
        backup_filename = get_backup_filename(original_filename)

        if not os.path.isfile(backup_filename):
            try:
                shutil.copy(original_filename, backup_filename)
            except FileNotFoundError:
                pass

    @staticmethod
    def is_configuration_restorable():
        return os.path.isfile(get_backup_filename(settings.get_interfaces_configuration_filename()))

    @staticmethod
    def restore_configuration(reboot=True, reboot_delay_seconds=5):
        if settings.is_network_configuration_enabled():
            original_filename = settings.get_interfaces_configuration_filename()
            backup_filename = get_backup_filename(settings.get_interfaces_configuration_filename())
            if os.path.isfile(backup_filename):
                try:
                    os.replace(backup_filename, original_filename)
                except Exception as ex:
                    logger.warning('Could not restore network configuration: %s', ex)
                else:
                    if reboot:
                        control_manager.reboot(delay_seconds=reboot_delay_seconds)
            else:
                logger.debug('Backup file %s was not found', backup_filename)
        else:
            logger.info('Network configuration change is disabled (for developer protection)')

    def get_configuration(self, interface):
        # TODO(dmu) MEDIUM: Implement more precise DHCP detection
        interface_stanza = get_stanza_by_interface(interface)
        if interface_stanza:
            match = IFACE_STANZA_RE.match(interface_stanza[0])
            is_dhcp = match and match.group('method') == 'dhcp'

            attributes = get_stanza_attributes(interface_stanza)
            attributes['dhcp'] = is_dhcp

            name_servers = attributes.get('name_servers')
            if name_servers is not None:
                attributes['name_servers'] = ','.join(ANY_SPACE_RE.split(name_servers.strip()))

            return attributes
        else:
            # We consider DHCP if interface is not configured in /etc/network/interfaces
            return {'dhcp': True}

    def is_network_configuration_supported(self):
        return (settings.is_vendored_hardware() and
                not is_inside_docker_container() and
                get_os_id_and_version() in OPERATING_SYSTEM_TEMPLATES)

    def get_configurable_network_interface(self):
        explicit_interface = settings.get_configurable_network_interface()
        if explicit_interface:
            return explicit_interface

        interfaces = get_interfaces()
        interfaces.remove('lo')  # do not consider local interface
        candidates = list(filter(lambda x: ':' not in x, interfaces))  # filter virtual interfaces
        if len(candidates) > 1:
            actual_ip_address = get_actual_ip_address()
            for interface in candidates.copy():
                ifaddresses = netifaces.ifaddresses(interface)
                for ipv4_address in ifaddresses.get(netifaces.AF_INET) or ():
                    if ipv4_address.get('addr') == actual_ip_address:
                        break
                else:
                    candidates.remove(interface)

        if candidates:
            if len(candidates) == 1:
                return candidates[0]
            else:
                logger.warning('Could not detect configurable network interface')
        else:
            logger.debug('No configurable network interface found')

    def set_network(self, interface, is_dhcp, context=None, reboot_delay_seconds=5,
                    no_backup=False):
        if not self.is_network_configuration_supported():
            raise NotImplementedError('Network configuration is not supported for this platform')

        replacing_content = get_replacing_content(interface, is_dhcp, context=context)
        stream = get_interfaces_configuration_file_stream(interface, replacing_content)

        if settings.is_network_configuration_enabled():
            if not no_backup:
                self.backup_configuration()

            with atomic_write(settings.get_interfaces_configuration_filename(),
                              overwrite=True) as f:
                f.write(stream.read())

            if reboot_delay_seconds is not None:
                control_manager.reboot(delay_seconds=reboot_delay_seconds,
                                       change_worker_stopped_status=False)
            # TODO(dmu) MEDIUM: Improve to change IP address without reboot (this includes proper
            #                   of restart Worker, Console and Watchdog to use updated network
            #                   settings)
            # update_network(interface, delay_seconds=self.update_network_delay_seconds)
        else:
            logger.info('Network configuration change was disabled (for developer protection)')
            logger.debug('Generated interfaces configuration file content:\n%s', stream.read())

    def set_static_ip_address(self, interface, ip_address, subnet_mask, default_gateway,
                              name_servers=('8.8.8.8', '8.8.4.4'), reboot_delay_seconds=5):
        context = {
            'ip_address': ip_address,
            'subnet_mask': subnet_mask,
            'default_gateway': default_gateway,
            'name_servers': name_servers,
        }
        self.set_network(interface, is_dhcp=False, context=context,
                         reboot_delay_seconds=reboot_delay_seconds)

    def set_dhcp(self, interface=None, reboot_delay_seconds=5, no_backup=False):
        interface = interface or self.get_configurable_network_interface()
        self.set_network(interface, is_dhcp=True, reboot_delay_seconds=reboot_delay_seconds,
                         no_backup=no_backup)


# As long as manager is stateless object we do not need to care about greenlet-multitasking safety
network_manager = NetworkManager()



import json

import httpretty
import gevent
import pytest
from sqlalchemy import create_engine

from packy_agent.configuration.settings import settings
from packy_agent.domain_logic.models.sqlalchemy.base import Session
from packy_agent.domain_logic.models.sqlalchemy.measurement import Measurement
from packy_agent.tests.worker.loops.main.base import start_main_loop, stop_main_loop


TASKS = [
    {
        'module_name': 'ping', 'args': [], 'id': 1,
        'kwargs': {
            'number_of_pings': 3,
            'host': '8.8.8.8',
            'interval_ms': 0,
            'packet_size': 64
        },
        'schedule': '*/2 * * * * *'
    },
]


@pytest.mark.usefixtures('activated')
def test_main_loop_produces_ping_measurements():
    main_loop = start_main_loop(TASKS)
    gevent.sleep(2)

    engine = create_engine(settings.get_database_url())
    with Session(engine) as session:
        measurements = session.query(Measurement).all()

    assert measurements
    for measurement in measurements:
        assert measurement.id
        assert measurement.measurement_type == 1
        assert measurement.created_at_ts
        assert not measurement.submitted_at_ts

        value_dict = json.loads(measurement.value)
        assert isinstance(value_dict['ts'], int)
        assert value_dict['target'] == '8.8.8.8'
        assert value_dict['n_pings'] == 3
        assert value_dict['packet_size'] == 64
        assert isinstance(value_dict['values'], list)
        if value_dict['values']:
            assert isinstance(value_dict['values'][0], (float, int))

    for task in main_loop.tasks:
        task.stop()

    gevent.sleep(2)

    with httpretty.enabled():
        url = settings.get_server_base_url() + 'api/v2/agent/module/ping/measurement/'
        httpretty.register_uri(httpretty.POST, url, status=201)

        gevent.sleep(3)

        latest_requests = httpretty.HTTPretty.latest_requests
        assert latest_requests
        measurements_submitted = 0
        for request in latest_requests:
            if request.path == '/api/v2/agent/module/ping/measurement/':
                body = json.loads(request.body)
                assert isinstance(body['ts'], int)
                assert body['target'] == '8.8.8.8'
                assert body['n_pings'] == 3
                assert body['packet_size'] == 64
                assert isinstance(body['values'], list)
                if body['values']:
                    assert all(isinstance(value, (float, int)) for value in body['values'])

                measurements_submitted += 1

        assert measurements_submitted > 0

    with Session(engine) as session:
        not_submitted = session.query(Measurement).filter_by(submitted_at_ts=None).all()
        assert not not_submitted

    stop_main_loop(main_loop)



import logging

from packy_agent.constants import PING_MODULE, TRACE_MODULE, SPEEDTEST_MODULE, HTTP_MODULE
from packy_agent.configuration.settings import settings
from packy_agent.worker.loops.task_results.consumer import TaskResultsConsumer
from packy_agent.worker.loops.task_results.submitter import TaskResultsSubmitter
from packy_agent.worker.loops.task_results.purger import TaskResultsPurger
from packy_agent.worker.loops.management.communication import PackyCommunicationLoop
from packy_agent.worker.loops.management.heartbeat import HeartbeatLoop
from packy_agent.worker.loops.management.guard import GuardLoop
from packy_agent.worker.loops.management.status import StatusLoop
from packy_agent.worker.loops.modules.ping import PingTaskLoop
from packy_agent.worker.loops.modules.traceroute import TracerouteTaskLoop
from packy_agent.worker.loops.modules.speedtest import SpeedtestTaskLoop
from packy_agent.worker.loops.modules.http import HTTPTaskLoop


MODULE_NAME_TO_LOOP_MAP = {
    PING_MODULE: PingTaskLoop,
    TRACE_MODULE: TracerouteTaskLoop,
    SPEEDTEST_MODULE: SpeedtestTaskLoop,
    HTTP_MODULE: HTTPTaskLoop,
}

logger = logging.getLogger(__name__)


class MainLoop:

    def __init__(self, task_definitions=None):
        # TODO(dmu) HIGH: We should not get task at initialization only. Instead we should
        #                 track changes (we have Packy Communication event for this) and add/remove
        #                 them on fly
        self.task_definitions = task_definitions or settings.get_tasks() or ()

        self.is_running = False
        self.is_stopping = False
        self.tasks = []

        self.task_results_consumer = None
        self.task_results_submitter = None
        self.task_results_purger = None

        self.communication_loop = None
        self.heartbeat_loop = None
        self.guard_loop = None
        self.status_loop = None

        self.loops = []

    def _start_loop(self, loop):
        if loop:
            loop.start()
            self.loops.append(loop)

    def _join_loops(self):
        for loop in self.loops:
            loop.join()

    def _stop_loops(self):
        for loop in self.loops:
            loop.stop()

    def run(self):
        if self.is_running:
            return

        self.is_running = True
        self.is_stopping = False
        logger.debug('STARTED main loop')

        active_loops = settings.get_worker_loops()

        if TaskResultsConsumer.formal_name in active_loops:
            self.task_results_consumer = TaskResultsConsumer()

        if TaskResultsSubmitter.formal_name in active_loops:
            self.task_results_submitter = TaskResultsSubmitter()

        if TaskResultsPurger.formal_name in active_loops:
            self.task_results_purger = TaskResultsPurger()

        if PackyCommunicationLoop.formal_name in active_loops:
            self.communication_loop = PackyCommunicationLoop(
                period=settings.get_worker_packy_communication_retry_period())

        if HeartbeatLoop.formal_name in active_loops:
            self.heartbeat_loop = HeartbeatLoop(
                period=settings.get_worker_heartbeat_period_seconds())

        if GuardLoop.formal_name in active_loops:
            self.guard_loop = GuardLoop(period=settings.get_worker_guard_period_seconds())

        if StatusLoop.formal_name in active_loops:
            self.status_loop = StatusLoop(self,
                                          address=settings.get_worker_http_bind_address(),
                                          port=settings.get_worker_http_port())

        if self.task_results_consumer:
            inbound_queue = self.task_results_consumer.inbound_queue
        else:
            inbound_queue = None

        tasks = []
        for td in self.task_definitions:
            module_name = td['module_name']
            loop = MODULE_NAME_TO_LOOP_MAP.get(module_name)
            if not loop:
                logger.debug('Unsupported module: %s', module_name)
                continue

            if loop.formal_name not in active_loops:
                continue

            tasks.append(
                loop(schedule=td['schedule'], args=td.get('args') or (),
                     kwargs=td.get('kwargs') or {},
                     outbound_queue=inbound_queue))

        self.tasks = tasks

        for task in self.tasks:
            self._start_loop(task)
        self._start_loop(self.task_results_consumer)
        self._start_loop(self.task_results_submitter)
        self._start_loop(self.task_results_purger)
        self._start_loop(self.communication_loop)
        self._start_loop(self.heartbeat_loop)
        self._start_loop(self.guard_loop)
        self._start_loop(self.status_loop)

        self._join_loops()

        self.is_running = False
        self.is_stopping = False
        logger.debug('STOPPED main loop')

    def stop(self):
        logger.debug('STOP requested for main loop')
        self.is_stopping = True
        self._stop_loops()



import logging

from packy_agent.exceptions import ImproperlyConfiguredError
from packy_agent.worker.loops.base.scheduled import ScheduledLoop
from packy_agent.worker.loops.base.producer import ProducerLoop


logger = logging.getLogger(__name__)


def get_target(kwargs, module_name):
    target = kwargs.get('target') or kwargs.get('host')
    if not target:
        logger.debug('Target (host) is not configured for %s', module_name)
        return

    return target


def get_packet_size(kwargs, module_name, default=64):
    packet_size = kwargs.get('packet_size')
    if packet_size is None:
        packet_size = default
        logger.warning('Packet size (packet_size) keyword argument was not provided, using '
                       '%s as default (%s)', packet_size, module_name)

    return packet_size


def get_number_of_pings(kwargs, module_name, default=10):
    number_of_pings = kwargs.get('number_of_pings')
    if number_of_pings is None:
        number_of_pings = default
        logger.warning('Number of pings (number_of_pings) keyword argument was not provided, '
                       'using %s as default (%s)', number_of_pings, module_name)

    return number_of_pings


class ScheduledProducerTaskLoop(ProducerLoop, ScheduledLoop):
    loop_type = 'scheduled_producer'



from schematics import types

from packy_agent.domain_logic.models.schematics.base import BaseMeasurement, CustomSchematicsModel


class Hop(CustomSchematicsModel):
    number = types.IntType()

    reply_number = types.IntType()
    ip_address = types.IPv4Type()

    sent = types.IntType(required=True)
    loss_abs = types.IntType(required=True)

    last = types.FloatType()
    best = types.FloatType()
    worst = types.FloatType()
    average = types.FloatType()
    stdev = types.FloatType()


class TraceModuleMeasurement(BaseMeasurement):
    target = types.StringType(required=True)
    packet_size = types.IntType(required=True)
    pings = types.IntType(required=True)
    hops = types.ListType(types.ModelType(Hop))



# Having it in a separate file to be able to access from Ansible scripts without unnecessary imports
# Please, be Python 2 compatible

defaults = {

    # Paths
    'settings_filename': '/etc/packy-agent-settings.yaml',
    'database_filename': None,
    'upgrade_in_progress_lock_filename': '/tmp/packy-upgrade-in-progress.lock',
    'packy_run_install_script_filename': '/tmp/packy-run-install-script.sh',
    'interfaces_configuration_filename': '/etc/network/interfaces',
    'configurable_network_interface': None,

    # URLs
    'server_base_url': None,
    'communication_url': None,
    'sentry_dsn': None,

    # Names
    'cached_settings_tablename': 'cached_settings',
    'local_settings_tablename': 'local_settings',

    # Flags
    'debug_mode': False,
    'activated': False,
    'reboot_enabled': True,
    'upgrade_enabled': True,
    'network_configuration_enabled': True,

    # Auth
    'client_id': None,
    'access_token': None,
    'refresh_token': None,
    'agent_key': None,
    'agent_name': None,

    # Timeouts/periods
    'http_timeout_seconds': 5,
    'process_status_check_period_seconds': 1,
    'process_stop_timeout_seconds': 10,
    'process_restart_timeout_seconds': 10,
    'sigterm_timeout_seconds': 10,

    # Misc
    'component': None,
    'minimum_disk_space_bytes': 1024 * 1024 * 512,  # 512Mb
    'vendored_hardware': False,
    'upgrade_max_duration_seconds': 60 * 60,  # 1 hour
    'restart_min_period_seconds': 120,
    'reboot_min_period_seconds': 300,

    'communication_roles': {
        'roles': {
            'subscriber': {},
            'publisher': {},
            'callee': {
                'shared_registration': True,
            },
            'caller': {},
        },
        'authmethods': ['ticket'],
        'authid': '*'
    },

    # Supervisor
    'supervisor': {
        'configuration_filename': '/etc/packy-supervisord.conf',
        'log_filename': '/var/log/packy/supervisor/supervisor.log',
        'run_directory': '/var/run/packy/supervisor',
    },

    # Console
    'console': {
        'flask': {},
        'uwsgi': {
            'configuration_filename': '/etc/packy-uwsgi.conf',
            'pid_filename': 'var/run/packy-uwsgi.pid',
        },
        'nginx': {
            'configuration_filename': '/etc/nginx/sites-enabled/packy-agent.conf'
        },
        'http_port': 8001,
        'stdout_log_filename': '/var/log/packy/console/standard.log',
        'stderr_log_filename': '/var/log/packy/console/error.log',
        'version_dump_filename': '/etc/packy-agent-console-version',
        'version_dump_variable_name': 'PACKY_AGENT_CONSOLE_VERSION',
    },

    # Worker
    'worker': {
        # General
        'pid': None,  # TODO(dmu) MEDIUM: should we rather use /var/run/packy-agent-worker.pid ?
        'stopped': False,
        'loops': ['ping_module', 'trace_module', 'speedtest_module', 'http_module', 'consumer',
                  'submitter', 'purger', 'communication', 'heartbeat', 'guard', 'status'],

        # Heartbeat
        'heartbeat_enabled': True,
        'heartbeat_period_seconds': 5,

        # Consumer
        'consumer_loop_timeout': 1,

        # Submitter
        'results_submission_period_seconds': 10,
        'results_submission_pause_seconds': 0.5,

        # Purger
        'purge_period_seconds': 300,
        'client_side_failed_submissions_timeout_seconds': 3600,

        # Paths
        'stdout_log_filename': '/var/log/packy/worker/standard.log',
        'stderr_log_filename': '/var/log/packy/worker/error.log',
        'version_dump_filename': '/etc/packy-agent-worker-version',
        'version_dump_variable_name': 'PACKY_AGENT_WORKER_VERSION',

        # Communication
        'packy_communication_retry_period': 30,
        'packy_communication_heartbeat_timeout_seconds': 60,

        # Watchdog related
        'offline_to_restart_seconds': 120,
        'restart_wait_seconds': 120,
        'offline_to_reboot_seconds': 300,
        'reboot_wait_seconds': 300,

        # Status HTTP server
        'http_bind_address': '127.0.0.1',
        'http_port': 5000,
        'expected_http_response_time_seconds': 1,  # for health check from Watchdog

        # Misc
        'activation_notify_period_seconds': 10,  # Server throttles anything more than 10 r/minute
        'started_check_period_seconds': 5,
        'periodic_task_jitter_factor': 0.1,
        'guard_period_seconds': 5,
    },

    # Watchdog
    'watchdog': {
        'check_period_seconds': 30,
        'relax_period_seconds': 60,
        'warmup_period_seconds': 30,

        'warning_report_period': 600,
        'error_report_period': 300,

        'stdout_log_filename': '/var/log/packy/watchdog/standard.log',
        'stderr_log_filename': '/var/log/packy/watchdog/error.log',
        'version_dump_filename': '/etc/packy-agent-watchdog-version',
        'version_dump_variable_name': 'PACKY_AGENT_WATCHDOG_VERSION',
    },

    # logging
    'log_level': 'WARNING',
    'logging': {
        'version': 1,
        'disable_existing_loggers': False,
        'formatters': {
            'standard': {
                'format': '%(asctime)s %(levelname)s %(name)s %(message)s'
            },
        },
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'standard',
            },
        },
        'loggers': {
            'packy_agent': {
                'propagate': True,
            },
            'sqlitedict': {
                'level': 'WARNING',
                'propagate': True,
            },
            'urllib3':  {
                'level': 'WARNING',
                'propagate': True,
            },
            'sentry_sdk':  {
                'level': 'WARNING',
                'propagate': True,
            },
        },
        'root': {
            'level': 'WARNING',
            'handlers': ['console'],
        }
    }
}



from urllib.parse import urlparse

import socket
import warnings

import netifaces

from packy_agent.configuration.settings import settings

DEFAULT_HOST_TO_REACH = '10.255.255.255'


def get_machine_ip_address(host_to_reach=DEFAULT_HOST_TO_REACH):
    # As suggested here:
    # https://stackoverflow.com/questions/166506/finding-local-ip-addresses-using-pythons-stdlib

    # TODO(dmu) HIGH: Support multiple interfaces properly (how is it?)
    socket_obj = None
    try:
        socket_obj = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        try:
            # to report the interface that is actually used for connection to Packy Server
            socket_obj.connect((host_to_reach, 1))
        except Exception:
            # destination does not have to be reachable
            socket_obj.connect((DEFAULT_HOST_TO_REACH, 1))
        return socket_obj.getsockname()[0]
    except Exception:
        warnings.warn('Could not get actual IP address')
        return
    finally:
        if socket_obj:
            socket_obj.close()


def get_hostname_from_url(url):
    return urlparse(url).netloc.split(':')[0]


def get_interfaces():
    return netifaces.interfaces()


def get_actual_ip_address():
    return get_machine_ip_address(get_hostname_from_url(settings.get_server_base_url()))



import os
from pkg_resources import resource_string, resource_filename


def get_package_file_content(module_name, filename):
    return resource_string(module_name, filename)


def get_package_file_full_path(module_name, filename):
    return os.path.abspath(resource_filename(module_name, filename))



import time
import logging
from collections import defaultdict, OrderedDict
from threading import Condition

from psutil import Process, NoSuchProcess
import requests
from requests.exceptions import ConnectionError as RequestsConnectionError, HTTPError
from sentry_sdk import capture_message, capture_exception

from packy_agent.configuration.settings import settings
from packy_agent.domain_logic.constants import (
    WORKER_COMPONENT, CONSOLE_COMPONENT, WATCHDOG_COMPONENT)
from packy_agent.clients.packy_server import get_packy_server_client
from packy_agent.clients.packy_agent_worker import get_packy_agent_worker_client
from packy_agent.constants import MODULE_LOOP_SUFFIX
from packy_agent.domain_logic.managers.control import control_manager
from packy_agent.utils.datetime import get_croniter
from packy_agent.domain_logic.managers.control import NGINX_SUPERVISOR_PROGRAM as NGINX
from packy_agent.utils.fs import get_available_disk_space


logger = logging.getLogger(__name__)


DAY_SECONDS = 60 * 60 * 24
MESSAGE_KEY_LEN = 10
MAX_THROTTLE_DICT_SIZE = 1024
FAILOVER_CHECK_PERIOD_SECONDS = 30
ROOT_PID = 1
START = 'start'
RESTART = 'restart'
REBOOT = 'reboot'
ERROR = 'error'
WARNING = 'warning'
UNKNOWN = 'unknown'
OK = 'ok'


def get_loop_deadline(loop_props):
    now_ts = time.time()
    loop_type = loop_props.get('loop_type')
    if loop_type in ('scheduled', 'scheduled_producer'):
        schedule = loop_props.get('schedule')
        if schedule:
            croniter = get_croniter(schedule)
            return croniter.get_next()
    elif loop_type == 'periodic':
        period = loop_props.get('period')
        if period is not None:
            return now_ts + period


def is_uptime_gte(component, duration):
    started_at_ts = settings.get_subkey(component, 'started_at_ts')
    return started_at_ts is not None and time.time() - started_at_ts >= duration


class WatchdogService:

    def __init__(self, raven_client=None):
        self.raven_client = raven_client

        self._graceful_stop = False
        self.condition = Condition()

        self.known_activated_time = None
        self.last_restart_time = None
        self.last_known_online_time = None

        self.last_known_online_ts = None

        self.has_reported_not_activated = False
        self.has_reported_stopped = False

        self.collected_counter = None
        self.collected_counter_ts = None
        self.collected_counter_deadline_ts = None
        self.submitted_counter = None
        self.submitted_counter_ts = None
        self.submitted_counter_deadline_ts = None
        self.purged_records = None
        self.purged_records_ts = None

        self.loop_stats = defaultdict(dict)

        self.report_throttle = OrderedDict()

    def graceful_stop(self):
        self._graceful_stop = True
        with self.condition:
            self.condition.notify_all()

    def throttle(self, message, period):
        now_ts = time.time()

        message_key = message[:MESSAGE_KEY_LEN]
        last_reported_ts = self.report_throttle.get(message_key)
        if last_reported_ts and time.time() < last_reported_ts + period:
            logger.debug('Message has been throttled: %s', message)
            return True
        else:
            self.report_throttle[message_key] = now_ts
            if len(self.report_throttle) > MAX_THROTTLE_DICT_SIZE:
                self.report_throttle.popitem(last=False)
            return False

    def report_warning(self, message):
        logger.warning(message)
        if not self.throttle(message, settings.get_watchdog_warning_report_period()):
            capture_message(message, level='warning')

    def report_error(self, message):
        logger.error(message)
        if not self.throttle(message, settings.get_watchdog_error_report_period()):
            capture_message(message, level='error')

    def start_worker(self):
        self.report_warning('Watchdog requested Worker start')
        control_manager.start(WORKER_COMPONENT)

    def restart_component(self, component):
        restart_min_period_seconds = settings.get_restart_min_period_seconds()
        restarted_at_ts = settings.get_subkey(component, 'restarted_at_ts')
        if (restarted_at_ts is not None and
                time.time() - restarted_at_ts < restart_min_period_seconds):
            logger.debug(f'Restart is requested for {component.capitalize()} again too soon')
        else:
            self.report_warning(f'Watchdog requested {component.capitalize()} restart')
            control_manager.restart(component)

            # TODO(dmu) LOW: Should we restart nginx inside `control_manager.restart()`?
            if component == CONSOLE_COMPONENT:
                control_manager.restart(NGINX)

    def restart_console(self):
        self.restart_component(CONSOLE_COMPONENT)

    def restart_worker(self):
        self.restart_component(WORKER_COMPONENT)

    def restart_watchdog(self):
        self.restart_component(WATCHDOG_COMPONENT)

    def reboot(self):
        reboot_min_period_seconds = settings.get_reboot_min_period_seconds()
        rebooted_at_ts = settings.get_rebooted_at_ts()
        if (rebooted_at_ts is not None and
                time.time() - rebooted_at_ts < reboot_min_period_seconds):
            logger.debug('Reboot is requested again too soon')
        else:
            self.report_warning('Watchdog requested reboot')
            control_manager.reboot()

    def resolve_online_status(self):
        logger.debug('Resolving agent online status...')
        now_ts = time.time()

        if not settings.get_worker_heartbeat_enabled():
            logger.debug('Heartbeat is disabled, so we do not expect agent to be online')
            return

        try:
            is_online = get_packy_server_client().is_agent_online()
        except Exception:
            logger.debug('Could not get agent online status. Maybe Packy Server is not available')
            return

        if is_online is None:
            logger.debug('Could not get agent online status. Maybe there is an issue with '
                         'Packy Server')
            return
        elif is_online:
            self.last_known_online_ts = now_ts

        if self.last_known_online_ts is None:
            return
        else:
            offline_period_seconds = now_ts - self.last_known_online_ts

            # If we have been offline long enough and rebooted long ago then go reboot
            if (offline_period_seconds >= settings.get_worker_offline_to_reboot_seconds() and
                    now_ts >= (settings.get_rebooted_at_ts() or 0) +
                               settings.get_worker_reboot_wait_seconds()):
                return REBOOT
            # If we have been offline long enough and started long ago then go restart
            elif (offline_period_seconds >= settings.get_worker_offline_to_restart_seconds() and
                  now_ts >= (max(settings.get_worker_started_at_ts() or 0,
                                 settings.get_worker_restarted_at_ts() or 0)) +
                             settings.get_worker_restart_wait_seconds()):
                return RESTART

    def resolve_worker_supervisor_program_status(self):
        logger.debug('Resolving Worker running status (supervisor)...')
        if control_manager.is_running(WORKER_COMPONENT):
            logger.debug('Worker is running (supervisor)')
            return
        elif not (control_manager.is_starting(WORKER_COMPONENT) or
                  control_manager.is_stopping(WORKER_COMPONENT)):
            logger.debug('Worker is NOT running (supervisor)')
            return START

    def resolve_worker_process_status(self):
        pid = control_manager.get_pid(WORKER_COMPONENT)

        try:
            Process(pid)
            return
        except NoSuchProcess:
            return RESTART

    def resolve_loops_activity(self, loops):
        now_ts = time.time()
        relax_period = settings.get_watchdog_relax_period_seconds()

        for loop_name, loop_props in loops.items():
            new_counter = loop_props.get('counter')
            current_counter = (self.loop_stats.get(loop_name) or {}).get('counter')
            if new_counter == current_counter:
                change_deadline_ts = self.loop_stats[loop_name].get('change_deadline_ts')
                if change_deadline_ts is not None:
                    relaxed_change_deadline_ts = change_deadline_ts + relax_period
                    if now_ts >= relaxed_change_deadline_ts:
                        counter_ts = self.loop_stats[loop_name].get('counter_ts')
                        self.report_warning(
                            f'{loop_name} counter is stale on {current_counter} at '
                            f'{counter_ts:.3f} (change expected before '
                            f'{relaxed_change_deadline_ts:.3f})')
                        self.loop_stats[loop_name][
                            'change_deadline_ts'] = get_loop_deadline(loop_props)
            else:
                self.loop_stats[loop_name]['counter'] = new_counter
                self.loop_stats[loop_name]['counter_ts'] = now_ts

                logger.debug('%s updated counter from %s to %s at %.3f', loop_name,
                             current_counter, new_counter, now_ts)

                self.loop_stats[loop_name][
                    'change_deadline_ts'] = get_loop_deadline(loop_props)

    def resolve_workflow(self, status):

        # Filter out measurement loops
        running_measurement_loops = {k: v for k, v in (status.get('loops') or {}).items()
                                     if k.endswith(MODULE_LOOP_SUFFIX)}
        if not running_measurement_loops:
            logger.debug('No active measurement loops found')
            return  # No measurement loops, so we should not expect counters increment

        deadline_ts = min(
            list(filter(None, (get_loop_deadline(loop_props) for loop_name, loop_props
                               in running_measurement_loops.items()))) or (None,))
        logger.debug('Workflow deadline is calculated as %s', deadline_ts)

        relax_period = settings.get_watchdog_relax_period_seconds()
        now_ts = time.time()

        collected_counter = status.get('collected_counter')
        if self.collected_counter == collected_counter:
            logger.debug('Collected counter did not change (value: %s)', self.collected_counter)
            if (self.collected_counter_deadline_ts and
                    now_ts >= (self.collected_counter_deadline_ts + relax_period)):
                self.report_warning(f'Measurement collection is stale on {self.collected_counter} '
                                    f'at {self.collected_counter_ts:.3f}')
                self.collected_counter_deadline_ts = deadline_ts
        else:
            self.collected_counter = collected_counter
            self.collected_counter_ts = now_ts
            self.collected_counter_deadline_ts = deadline_ts

        submitted_counter = status.get('submitted_counter')
        if self.submitted_counter == submitted_counter:
            logger.debug('Submitted counter did not change (value: %s)', self.submitted_counter)
            if (self.submitted_counter_deadline_ts and
                    now_ts >= self.submitted_counter_deadline_ts + relax_period):
                self.report_warning(f'Measurement submission is stale on {self.submitted_counter} '
                                    f'at {self.submitted_counter_ts:.3f}')
                self.submitted_counter_deadline_ts = deadline_ts
        else:
            self.submitted_counter = submitted_counter
            self.submitted_counter_ts = now_ts
            self.submitted_counter_deadline_ts = deadline_ts

        purged_records = status.get('purged_records')
        if self.purged_records == purged_records:
            purge_period = settings.get_worker_purge_period_seconds() + relax_period
            if self.purged_records_ts and now_ts >= self.purged_records_ts + purge_period:
                self.report_warning(f'Measurement purging is stale on {self.purged_records} '
                                    f'at {self.purged_records_ts:.3f}')

        else:
            self.purged_records = purged_records
            self.purged_records_ts = now_ts

    def resolve_worker_response(self):
        start = time.time()
        try:
            status = get_packy_agent_worker_client().get_status()
            finish = time.time()
        except RequestsConnectionError:
            logger.debug('Could not get Worker status over HTTP')
            return RESTART
        except HTTPError as ex:
            self.report_error(f'Packy Agent Worker returned HTTP{ex.response.status_code}: {ex!r}')
            return

        actual_duration = finish - start
        expected_duration = settings.get_worker_expected_http_response_time_seconds()
        if actual_duration > expected_duration:
            self.report_warning(f'It took {actual_duration:.3g} seconds for Packy Agent Worker '
                                f'to respond while it was expected to take only '
                                f'{expected_duration:.3g}')

        loops = status.get('loops')
        if loops:
            self.resolve_loops_activity(loops)
        else:
            self.report_warning('Worker did not return running loops')

        self.resolve_workflow(status)

    def handle_console(self):
        if is_uptime_gte(CONSOLE_COMPONENT, DAY_SECONDS):
            self.restart_console()
            return

        console_base_url = settings.get_console_base_url()
        try:
            response = requests.get(console_base_url)
        except RequestsConnectionError:
            self.restart_console()
            return
        else:
            if response.status_code != 200:
                self.restart_console()
                return

    def get_worker_action(self):
        if not settings.is_activated():
            if not self.has_reported_not_activated:
                self.has_reported_not_activated = True
                logger.info('Worker is not activated')

            return
        self.has_reported_not_activated = False

        if settings.is_worker_stopped():
            if not self.has_reported_stopped:
                self.has_reported_stopped = True
                logger.info('Worker is stopped on purpose')

            return
        self.has_reported_stopped = False

        for method in (self.resolve_online_status, self.resolve_worker_supervisor_program_status,
                       self.resolve_worker_process_status, self.resolve_worker_response):
            action = method()
            if action:
                return action

        if is_uptime_gte(WORKER_COMPONENT, DAY_SECONDS):
            return RESTART

    def handle_worker(self):
        action = self.get_worker_action()
        if action:
            logger.debug('Action is required for Worker: %s', action)
            if action == RESTART:
                self.restart_worker()
            elif action == START:
                self.start_worker()
            elif action == REBOOT:
                self.reboot()
        else:
            logger.debug('No action is required for worker')

    def handle_watchdog(self):
        if is_uptime_gte(WATCHDOG_COMPONENT, DAY_SECONDS):
            self.restart_watchdog()

    def handle_general(self):
        minimum_disk_space_bytes = settings.get_minimum_disk_space_bytes()
        available_disk_space = get_available_disk_space()
        if available_disk_space < minimum_disk_space_bytes:
            self.report_warning(f'Less than {minimum_disk_space_bytes} bytes of disk space is '
                                f'remaining: {available_disk_space} bytes')

    def loop_iteration(self):
        if (time.time() - settings.get_watchdog_started_at_ts() <
                settings.get_watchdog_warmup_period_seconds()):
            logger.debug('Warming up...')
            return

        try:
            self.handle_general()
        except Exception:
            logger.exception('Error during general handling')
            capture_exception()

        try:
            self.handle_worker()
        except Exception:
            logger.exception('Error while handling Worker')
            capture_exception()

        try:
            self.handle_console()
        except Exception:
            logger.exception('Error while handling Console')
            capture_exception()

        try:
            self.handle_watchdog()
        except Exception:
            logger.exception('Error while handling Watchdog')
            capture_exception()

    def run(self):
        logger.info('STARTED Watchdog')
        while not self._graceful_stop:
            start = time.time()
            try:
                self.loop_iteration()
            except Exception:
                logger.exception('Error during loop iteration')
                capture_exception()

            # Configuration may change with time, so we reread it every loop
            try:
                check_period_seconds = settings.get_watchdog_check_period_seconds()
            except Exception:
                check_period_seconds = FAILOVER_CHECK_PERIOD_SECONDS
                logger.exception('Could not get check period seconds from configuration file, '
                                 'using failover value: %s', check_period_seconds)
                capture_exception()

            wait_duration = check_period_seconds - (time.time() - start)
            if wait_duration > 0:
                logger.debug('Waiting for %.3g seconds for next iteration', wait_duration)
                with self.condition:
                    self.condition.wait(wait_duration)

        logger.info('STOPPED Watchdog gracefully')



import os
import sys

from packy_agent.domain_logic.managers.install_and_upgrade import install_and_upgrade_manager
from packy_agent.configuration.settings import settings
from packy_agent.utils.cli import get_base_argument_parser
from packy_agent.utils.logging import configure_logging, configure_logging_basic


ACTIVATE_SCRIPT_PATH = '/tmp/packy-activate.sh'
ACTIVATE_PACKY_AGENT_SH = """#!/usr/bin/env bash
source {venv_path}/bin/activate
packy-agent-activate
"""
DEFAULT_LOG_LEVEL = 'ERROR'


def entry():
    configure_logging_basic(DEFAULT_LOG_LEVEL)
    parser = get_base_argument_parser(__loader__.name, default_log_level=DEFAULT_LOG_LEVEL)
    args = parser.parse_args()
    settings.set_commandline_arguments(vars(args))
    configure_logging(DEFAULT_LOG_LEVEL)

    print()
    print('-' * 40)
    print()
    print('Welcome to Packy Agent')
    print()

    if not settings.is_activated():
        with open(ACTIVATE_SCRIPT_PATH, 'w') as f:
            f.write(ACTIVATE_PACKY_AGENT_SH.format(venv_path=os.getenv('VIRTUAL_ENV')))
        os.system(f'chmod u+x {ACTIVATE_SCRIPT_PATH}')

        url = settings.get_console_base_url()
        print(f'Please, activate your Packy Agent at {url} or run `sudo {ACTIVATE_SCRIPT_PATH}`')
        print()


if __name__ == '__main__':
    sys.exit(entry())



from abc import ABC

import gevent

from packy_agent.worker.loops.base.loop import Loop


class StartableLoop(Loop, ABC):

    loop_type = 'startable'

    def __init__(self, is_safe_iteration=True, is_logged_iteration=True):
        super().__init__(is_safe_iteration=is_safe_iteration,
                         is_logged_iteration=is_logged_iteration)
        self.greenlet = None
        self.is_started = False

    def get_greenlet(self):
        return self.greenlet

    def start(self):
        if self.is_started:
            return

        self.greenlet = gevent.Greenlet(self.loop)
        self.greenlet.start()
        self.is_started = True

    def join(self):
        self.greenlet.join()



import logging
from abc import ABC

from gevent.queue import Queue

from packy_agent.worker.loops.base.misc import CallableLoop

logger = logging.getLogger(__name__)


class ProducerLoop(CallableLoop, ABC):

    loop_type = 'producer'

    def __init__(self, callable_=None, args=(), kwargs=None, is_safe_iteration=True,
                 is_logged_iteration=True, outbound_queue=None, **mkwargs):
        super().__init__(callable_=callable_, args=args, kwargs=kwargs,
                         is_safe_iteration=is_safe_iteration,
                         is_logged_iteration=is_logged_iteration, **mkwargs)

        self.produced_counter = 0
        self.outbound_queue = outbound_queue

    def call(self):
        logger.info('STARTED actual work: %s', self.description)
        result = self.callable(*self.args, **self.kwargs)
        logger.info('FINISHED actual work: %s', self.description)
        if result is None:
            logger.debug('Did not get result for %s', self.description)
        else:
            if self.outbound_queue:
                self.outbound_queue.put(result)

            self.produced_counter += 1

        return result



import logging

import flask
from flask.views import MethodView

from packy_agent.domain_logic.managers.control import control_manager
from packy_agent.utils.auth import activation_and_authentication_required

logger = logging.getLogger(__name__)


class IndexView(MethodView):

    @activation_and_authentication_required
    def get(self):
        context = {
            'worker_status': control_manager.get_worker_status_line(),
            'active_menu_item': 'status',
        }
        return flask.render_template('index.html', **context)




import time
import os
import os.path
from contextlib import contextmanager

import yaml
import pytest
from alembic.config import Config
from alembic import command

from packy_agent.utils.pkg_resources import get_package_file_full_path
from packy_agent.configuration.settings import settings
from packy_agent.configuration.sources.base import MappingBackedSource

CONFIG_ENV_VAR_NAME = 'PACKY_AGENT_SETTINGS_FILENAME'


def prefix_filename(path, prefix='test-'):
    head, tail = os.path.split(path)
    return os.path.join(head, prefix + tail)


@pytest.fixture(scope='session', autouse=True)
def test_config():
    config_path = os.environ.get(CONFIG_ENV_VAR_NAME)
    if not config_path:
        raise Exception(f'Environment variable {CONFIG_ENV_VAR_NAME} is not defined')

    with open(config_path) as f:
        local_config = yaml.load(f)

    database_filename = local_config.get('database_filename')
    if not database_filename:
        raise Exception(f'database_filename key is not defined in {config_path}')

    test_config_path = prefix_filename(config_path)
    try:
        os.remove(test_config_path)
    except FileNotFoundError:
        pass

    # TODO(dmu) LOW: Maybe use `override_settings` instead?
    # Make timeouts smaller
    local_config['worker']['consumer_loop_timeout'] = 0.1
    local_config['worker']['results_submission_period_seconds'] = 0.1
    local_config['worker']['results_submission_pause_seconds'] = 0.1
    local_config['worker']['purge_period_seconds'] = 1
    local_config['worker']['loops'] = [
        'ping_module', 'trace_module', 'http_module', 'speedtest_module', 'consumer', 'submitter',
        'purger']

    # Substitute names
    local_config['database_filename'] = prefix_filename(database_filename)
    local_config['server_base_url'] = 'http://test'
    local_config['communication_url'] = 'ws://test/ws'
    local_config['sentry_dns'] = None

    with open(test_config_path, 'w') as f:
        yaml.dump(local_config, f, default_flow_style=False)

    os.environ[CONFIG_ENV_VAR_NAME] = test_config_path

    return test_config_path


@pytest.fixture(autouse=True)
def local_database(test_config):
    from packy_agent.configuration.settings import settings
    database_filename = settings.get_database_filename()
    # At this moment key-value storage tables are created
    try:
        os.remove(database_filename)
    except FileNotFoundError:
        pass

    # Recreate key-value storage tables
    settings._local_settings._mapping = None
    settings._cached_settings._mapping = None

    alembic_cfg = Config(get_package_file_full_path('packy_agent', 'configuration/alembic.ini'))
    command.upgrade(alembic_cfg, 'head')


@pytest.fixture(autouse=True)
def logging(test_config):
    from packy_agent.utils.logging import configure_logging
    configure_logging()


@pytest.fixture
def activated():
    from packy_agent.configuration.settings import settings

    settings.activate(
        client_id='client_id',
        access_token='access_token_value',
        refresh_token='refresh_token_value',
        agent_key='agent_key_value',
        agent_name='name_value',
    )


@contextmanager
def override_setting(key, value):
    settings._sources.insert(0, MappingBackedSource({key: value}))
    yield
    settings._sources.pop(0)


@contextmanager
def override_settings(new_settings):
    settings._sources.insert(0, MappingBackedSource(new_settings))
    yield
    settings._sources.pop(0)


@contextmanager
def local_settings(new_settings):
    from packy_agent.configuration.settings import settings
    backup = {}
    for k, v in new_settings.items():
        if k in settings._local_settings:
            backup[k] = settings._local_settings[k]

    # TODO(dmu) LOW: Implement deep update here
    settings._local_settings.update(new_settings)
    yield
    settings._local_settings.update(backup)


@pytest.fixture
def watchdog_started_long_ago():
    with override_settings({'watchdog': {'started_at_ts': time.time() - 3600}}):
        yield



from unittest.mock import MagicMock

import pytest

from packy_agent.watchdog.service import WatchdogService


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_loop_iteration():
    service = WatchdogService()
    service.loop_iteration()


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_all_worker_resolves_tried():
    service = WatchdogService()
    service.resolve_online_status = MagicMock(return_value=None)
    service.resolve_worker_supervisor_program_status = MagicMock(return_value=None)
    service.resolve_worker_process_status = MagicMock(return_value=None)
    service.resolve_worker_response = MagicMock(return_value=None)
    service.loop_iteration()

    service.resolve_online_status.assert_called()
    service.resolve_worker_supervisor_program_status.assert_called()
    service.resolve_worker_process_status.assert_called()
    service.resolve_worker_response.assert_called()



from schematics import types

from packy_agent.domain_logic.models.schematics.base import BaseMeasurement


class PingModuleMeasurement(BaseMeasurement):
    target = types.StringType(required=True)
    packet_size = types.IntType(required=True)
    n_pings = types.IntType(required=True)
    values_ = types.ListType(types.FloatType, serialized_name='values', deserialize_from='values')



import logging

import flask
from flask.views import MethodView
from flask import redirect, url_for, flash, render_template

from packy_agent.console.forms.activate import ActivateForm
from packy_agent.console.forms.activate_extra import ActivateExtraForm
from packy_agent.console.views.base import smart_redirect
from packy_agent.exceptions import AuthenticationError, ValidationError
from packy_agent.domain_logic.managers.install_and_upgrade import install_and_upgrade_manager
from packy_agent.utils.auth import set_authentication_cookie, not_activated
from packy_agent.clients.packy_server import get_inactive_agents

logger = logging.getLogger(__name__)


class ActivateView(MethodView):

    @not_activated
    def get(self):
        return flask.render_template('activate.html', form=ActivateForm())

    @not_activated
    def post(self):
        form = ActivateForm()
        if form.validate():
            email = form.email.data
            password = form.password.data

            try:
                agent_id = None
                agents = get_inactive_agents(email, password)
                if agents:
                    # There inactivate agents, so we have to ask if we should reactivate one of them
                    form = ActivateExtraForm(email=email, password=password)
                    agent_choices = tuple((str(id_), name) for id_, name in agents.items())
                    form.agent.choices += agent_choices

                    if form.validate():
                        if form.agent.data != 'new':
                            # User to reactivate one of the agents, let's take its id
                            agent_id = int(form.agent.data)
                            if agent_id not in agents:
                                flask.flash('Invalid agent')
                                return redirect(url_for('activation_failure'))
                    elif form.extra.data == 'yes':
                        # Data did not pass validation and we need to render errors on extra form
                        return render_template('activate_extra.html', form=form)
                    else:
                        # Data did not pass validation because `activate.html` was actually
                        # submitted, so return extra form with a list of inactive agents
                        form = ActivateExtraForm(email=email, password=password)
                        form.agent.choices += agent_choices
                        return render_template('activate_extra.html', form=form)

                # Otherwise reactivate or create or agent
                install_and_upgrade_manager.activate(email, password, agent_id=agent_id)

            except AuthenticationError:
                flash('Not authenticated (invalid credentials)')
                return flask.redirect(flask.url_for('activation_failure'))
            except ValidationError as ex:
                flash(str(ex))
                return flask.redirect(flask.url_for('activation_failure'))
            except Exception:
                logger.exception('Error during activation')
                flash('Error during activation')
                return redirect(flask.url_for('activation_failure'))

            set_authentication_cookie()
            return smart_redirect('success', 'index', button_text='Continue')

        return render_template('activate.html', form=form)


class ActivationFailureView(MethodView):

    def get(self):
        return render_template('activation_failure.html')



import logging

import flask
from flask.views import MethodView

from packy_agent.domain_logic.managers.install_and_upgrade import install_and_upgrade_manager
from packy_agent.console.forms.action import ActionForm
from packy_agent.configuration.settings import settings
from packy_agent.console.views.base import smart_redirect
from packy_agent.utils.auth import activation_and_authentication_required
from packy_agent.utils.installation import is_upgrade_in_progress


DELAY_SECONDS = 5

logger = logging.getLogger(__name__)


class UpgradeView(MethodView):

    @activation_and_authentication_required
    def get(self):
        form = ActionForm()
        context = {
            'form': form,
            'active_menu_item': 'upgrade',
        }

        return flask.render_template('upgrade.html', **context)

    @activation_and_authentication_required
    def post(self):
        form = ActionForm()
        if form.validate():
            action = flask.request.form['action']
            if action == 'upgrade_and_restart':
                if is_upgrade_in_progress():
                    flask.flash('Concurrent upgrade detected')
                    return smart_redirect('failure', 'upgrade')
                else:
                    flask.flash(
                        'Upgrade and restart will start in {} seconds...'.format(DELAY_SECONDS))
                    install_and_upgrade_manager.upgrade_and_restart(DELAY_SECONDS)
            else:
                raise ValueError('Unknown action: {}'.format(action))

            if not settings.is_upgrade_enabled():
                flask.flash('Please, see logs for more information')

            return smart_redirect('success', 'upgrade')

        return flask.render_template('upgrade.html', form=form)



import logging

import flask
from flask.views import MethodView

from packy_agent.domain_logic.managers.control import control_manager
from packy_agent.console.forms.action import ActionForm
from packy_agent.console.views.base import smart_redirect
from packy_agent.utils.auth import activation_and_authentication_required
from packy_agent.domain_logic.constants import WORKER_COMPONENT, CONSOLE_COMPONENT


DELAY_SECONDS = 5

logger = logging.getLogger(__name__)


def do_action(action):
    if action == 'reboot':
        flask.flash('Reboot will start in {} seconds...'.format(DELAY_SECONDS))
        control_manager.reboot(delay_seconds=DELAY_SECONDS)
    elif action == 'start_worker':
        control_manager.start(WORKER_COMPONENT)
    elif action == 'stop_worker':
        control_manager.stop(WORKER_COMPONENT)
    elif action == 'restart_worker':
        control_manager.restart(WORKER_COMPONENT)
    elif action == 'restart':
        # Restart is delayed because we need to send HTTP response before Control Server goes down
        flask.flash('Restart will start in {} seconds...'.format(DELAY_SECONDS))
        control_manager.restart_all(delay_seconds=DELAY_SECONDS)
    elif action == 'restart_console':
        # Restart is delayed because we need to send HTTP response before Control Server goes down
        flask.flash('Restart will start in {} seconds...'.format(DELAY_SECONDS))
        control_manager.restart(CONSOLE_COMPONENT, async_delay_seconds=DELAY_SECONDS)
    else:
        raise ValueError('Unknown action: {}'.format(action))


class ControlView(MethodView):

    @activation_and_authentication_required
    def get(self):
        context = {
            'form': ActionForm(),
            'worker_status': control_manager.get_worker_status_line(),
            'active_menu_item': 'control',
        }

        return flask.render_template('control.html', **context)

    @activation_and_authentication_required
    def post(self):
        form = ActionForm()
        if form.validate():
            action = flask.request.form['action']
            do_action(action)
            return smart_redirect('success', 'control')

        return flask.render_template('control.html', form=form)



import time
import logging

from packy_agent.worker.loops.base.periodic import PeriodicLoop
from packy_agent.clients.packy_communication import PackyCommunicationClient
from packy_agent.domain_logic.managers.control import control_manager
from packy_agent.domain_logic.managers.install_and_upgrade import install_and_upgrade_manager
from packy_agent.configuration.settings import settings
from packy_agent.constants import (COMMUNICATION_LOOP, RELOAD_EVENT, RESTART_EVENT, REBOOT_EVENT,
                                   UPGRADE_EVENT, HEARTBEAT_EVENT)

logger = logging.getLogger(__name__)


class PackyCommunicationLoop(PeriodicLoop):

    formal_name = COMMUNICATION_LOOP

    def __init__(self, period, is_logged_iteration=True, **mkwargs):
        super().__init__(period=period, is_logged_iteration=is_logged_iteration, **mkwargs)

        self.communication_client = None
        self.heartbeat_request_ts = None
        self.heartbeat_received_ts = None

        self.agent_key = settings.get_agent_key()

    def is_communication_timed_out(self):
        if self.heartbeat_request_ts is None:
            return False

        if self.heartbeat_received_ts is None:
            return (time.time() - self.heartbeat_request_ts >
                    settings.get_worker_packy_communication_heartbeat_timeout_seconds())

        return self.heartbeat_received_ts < self.heartbeat_request_ts

    def handle_reload(self, *args, **kwargs):
        # We are restarting everything because some settings are read only once on start and
        # we need to reread them
        logger.info('PROCESSING reload event...')
        control_manager.restart_all()

    def handle_restart(self, *args, **kwargs):
        # We are restarting everything because some settings are read only once on start and
        # we need to reread them
        logger.info('PROCESSING restart event...')
        control_manager.restart_all()

    def handle_reboot(self, *args, **kwargs):
        logger.info('PROCESSING reboot event...')
        control_manager.reboot()

    def handle_upgrade(self, *args, **kwargs):
        logger.info('PROCESSING upgrade event...')
        install_and_upgrade_manager.upgrade_and_restart()

    def handle_heartbeat(self, *args, **kwargs):
        logger.debug(f'PROCESSING heartbeat event ({args}, {kwargs})...')
        self.heartbeat_received_ts = time.time()

    def send_heartbeat(self):
        self.heartbeat_request_ts = self.communication_client.send_heartbeat(self.agent_key)

    def destroy_communication_client(self):
        if self.communication_client:
            self.communication_client.stop()
            self.communication_client = None

        self.heartbeat_request_ts = None
        self.heartbeat_received_ts = None

    def iteration(self):
        if self.communication_client:
            if self.is_communication_timed_out():
                logger.warning('Communication timed out')
                self.destroy_communication_client()
            else:
                self.send_heartbeat()
                return

        if not self.communication_client:  # True after self.destroy_communication_client()
            self.communication_client = PackyCommunicationClient()
            self.communication_client.set_handler(RELOAD_EVENT, self.handle_reload)
            self.communication_client.set_handler(RESTART_EVENT, self.handle_restart)
            self.communication_client.set_handler(REBOOT_EVENT, self.handle_reboot)
            self.communication_client.set_handler(UPGRADE_EVENT, self.handle_upgrade)
            self.communication_client.set_handler(HEARTBEAT_EVENT, self.handle_heartbeat)
            try:
                self.communication_client.start()
            except Exception as ex:
                logger.warning(f'Could not start {self.description}: {ex!r}')
                self.destroy_communication_client()

    def stop(self):
        super().stop()
        self.destroy_communication_client()



import logging

from requests.exceptions import ConnectionError, HTTPError
import flask
from flask.views import MethodView

from packy_agent.clients.packy_server import get_packy_server_client
from packy_agent.console.forms.login import LoginForm

from packy_agent.utils.auth import set_authentication_cookie, activation_required


logger = logging.getLogger(__name__)


def login_failure():
    return flask.redirect(flask.url_for('login_failure'))


class LoginView(MethodView):

    @activation_required
    def get(self):
        return flask.render_template('login.html', form=LoginForm())

    @activation_required
    def post(self):
        form = LoginForm()
        if form.validate():
            try:
                is_valid = get_packy_server_client().validate_auth(form.email.data, form.password.data)
            except ConnectionError:
                flask.flash('Packy Server is unavailable for authentication. '
                            'Please, try again later')
                return login_failure()
            except HTTPError as ex:
                status_code = ex.response.status_code
                if 400 <= status_code < 500:
                    flask.flash('An error occurred during authentication. '
                                'Please, upgrade Packy Agent and/or try later')
                    return login_failure()
                elif 500 <= status_code < 600:
                    flask.flash('Packy Server is failure during authentication. '
                                'Please, again try later')
                    return login_failure()
                else:
                    raise

            if not is_valid:
                flask.flash('Not authenticated (invalid credentials)')
                return login_failure()

            set_authentication_cookie()
            url = flask.request.args.get('next')
            if not url:
                url = flask.url_for('index')

            return flask.redirect(url)

        return flask.render_template('login.html', form=form)


class LoginFailureView(MethodView):

    def get(self):
        return flask.render_template('login_failure.html')



import logging
import logging.config
from packy_agent.configuration.settings import settings

DEFAULT_LOGGING_FORMAT = '%(asctime)s %(levelname)s %(name)s %(message)s'


logger = logging.getLogger(__name__)


def configure_logging_basic(level=logging.DEBUG):
    logging.basicConfig(format=DEFAULT_LOGGING_FORMAT, level=level)


def configure_logging(basic_level=logging.DEBUG):
    configure_logging_basic(level=basic_level)
    logger.info('Applied basic logging configuration first to debug logging configuration '
                'issues themselves')
    logging.config.dictConfig(settings.get_logging())
    logging.getLogger().setLevel(settings.get_log_level())
    logger.info('FINISHED applying logging configuration')



from packy_agent.configuration.sources.defaults import ConfigurationDefaults


def test_defaults_smoke():
    defaults = ConfigurationDefaults()
    assert 'log_level' in defaults
    assert defaults['log_level'] == 'WARNING'



"""empty message

Revision ID: fa9a67d72a78
Revises: 54f648ebdb64
Create Date: 2019-01-02 15:17:48.104469

"""
from alembic import op
import sqlalchemy as sa

# revision identifiers, used by Alembic.
revision = 'fa9a67d72a78'
down_revision = '54f648ebdb64'
branch_labels = None
depends_on = None


def upgrade():
    op.add_column('measurement', sa.Column('error_side', sa.Integer(), nullable=True))


def downgrade():
    op.drop_column('measurement', 'error_side')



import sys
import logging

from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm.session import Session as SQLAlchemySession, make_transient


logger = logging.getLogger(__name__)


class Session(SQLAlchemySession):

    def __init__(self, bind=None, **kwargs):
        kwargs.setdefault('autocommit', False)
        # `expire_on_commit` is needed to avoid db queries when accessing model instance
        # attributes after session commit
        kwargs.setdefault('expire_on_commit', False)
        super(Session, self).__init__(bind=bind, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.close_commit_or_rollback(exc_type, exc_value, exc_traceback)

    def close_commit_or_rollback(self, exc_type=None, exc_value=None, exc_traceback=None):
        session_objects = list(self)

        commit_exc_type = None
        commit_exc_value = None
        commit_exc_traceback = None
        if exc_type:
            logger.exception('Rolling back after exception:',
                             exc_info=(exc_type, exc_value, exc_traceback))
            self.expunge_all()  # Remove all objects from session (must be called before rollback)
            self.rollback()
            # No need to raise exception explicitly it will be "reraised" implicitly
            # on return from __exit__()
        else:
            try:
                self.commit()
            except:
                commit_exc_type, commit_exc_value, commit_exc_traceback = sys.exc_info()
                try:
                    logger.exception('Error while committing, rolling back',
                                     exc_info=(commit_exc_type, commit_exc_value,
                                               commit_exc_traceback))
                    self.expunge_all()
                    self.rollback()
                except:
                    commit_exc_type, commit_exc_value, commit_exc_traceback = sys.exc_info()
                    logger.exception('Error while rolling back after error while committing '
                                     '(we did our best to recover)',
                                     exc_info=(commit_exc_type, commit_exc_value,
                                               commit_exc_traceback))

        # Disconnect session objects from session completely (stronger than `expunge()`)
        for obj in session_objects:
            make_transient(obj)

        self.close()

        if commit_exc_type:
            raise commit_exc_value.with_traceback(commit_exc_traceback)


ModelBase = declarative_base()



from unittest.mock import MagicMock

import httpretty
import pytest

from packy_agent.watchdog.service import WatchdogService
from packy_agent.configuration.settings import settings
from packy_agent.tests.fixtures.configuration import override_settings


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_smoke():
    with httpretty.enabled():
        url = 'http://{}:{}/status/'.format(settings.get_worker_http_bind_address(),
                                            settings.get_worker_http_port())
        httpretty.register_uri(httpretty.GET, url, status=200, body='{"loops": {"loop1": {}}}')
        httpretty.register_uri(httpretty.GET, settings.get_console_base_url(),
                               status=200)  # emulate console running

        service = WatchdogService()

        service.reboot = MagicMock()
        service.restart_worker = MagicMock()
        service.start_worker = MagicMock()

        service.resolve_online_status = MagicMock()
        service.resolve_worker_supervisor_program_status = MagicMock()
        service.resolve_worker_process_status = MagicMock()
        service.resolve_loops_activity = MagicMock()
        service.resolve_workflow = MagicMock()

        service.loop_iteration()

        service.start_worker.assert_not_called()
        service.restart_worker.assert_not_called()
        service.reboot.assert_not_called()


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_not_responding():
    service = WatchdogService()

    service.reboot = MagicMock()
    service.restart_worker = MagicMock()
    service.start_worker = MagicMock()

    service.resolve_online_status = MagicMock(return_value=None)
    service.resolve_worker_supervisor_program_status = MagicMock(return_value=None)
    service.resolve_worker_process_status = MagicMock(return_value=None)
    service.resolve_loops_activity = MagicMock(return_value=None)
    service.resolve_workflow = MagicMock(return_value=None)

    service.loop_iteration()

    service.start_worker.assert_not_called()
    service.restart_worker.assert_called()
    service.reboot.assert_not_called()


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_http5xx_response():
    with httpretty.enabled():
        url = 'http://{}:{}/status/'.format(settings.get_worker_http_bind_address(),
                                            settings.get_worker_http_port())
        httpretty.register_uri(httpretty.GET, url, status=500)
        httpretty.register_uri(httpretty.GET, settings.get_console_base_url(),
                               status=200)  # emulate console running

        service = WatchdogService()

        service.reboot = MagicMock()
        service.restart_worker = MagicMock()
        service.start_worker = MagicMock()

        service.resolve_online_status = MagicMock(return_value=None)
        service.resolve_worker_supervisor_program_status = MagicMock(return_value=None)
        service.resolve_worker_process_status = MagicMock(return_value=None)
        service.resolve_loops_activity = MagicMock(return_value=None)
        service.resolve_workflow = MagicMock(return_value=None)
        service.report_error = MagicMock()

        service.loop_iteration()

        service.start_worker.assert_not_called()
        service.restart_worker.assert_not_called()
        service.reboot.assert_not_called()

        service.report_error.assert_called()
        call_args = service.report_error.call_args
        assert call_args
        assert call_args[0]
        assert call_args[0][0].startswith('Packy Agent Worker returned HTTP500:')


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_no_loops_in_response():
    with httpretty.enabled():
        url = 'http://{}:{}/status/'.format(settings.get_worker_http_bind_address(),
                                            settings.get_worker_http_port())
        httpretty.register_uri(httpretty.GET, url, status=200, body='{}')
        httpretty.register_uri(httpretty.GET, settings.get_console_base_url(),
                               status=200)  # emulate console running

        service = WatchdogService()

        service.reboot = MagicMock()
        service.restart_worker = MagicMock()
        service.start_worker = MagicMock()

        service.resolve_online_status = MagicMock(return_value=None)
        service.resolve_worker_supervisor_program_status = MagicMock(return_value=None)
        service.resolve_worker_process_status = MagicMock(return_value=None)
        service.resolve_loops_activity = MagicMock(return_value=None)
        service.resolve_workflow = MagicMock(return_value=None)
        service.report_error = MagicMock()
        service.report_warning = MagicMock()

        service.loop_iteration()

        service.start_worker.assert_not_called()
        service.restart_worker.assert_not_called()
        service.reboot.assert_not_called()
        service.resolve_loops_activity.assert_not_called()

        service.report_warning.assert_called()
        call_args = service.report_warning.call_args
        assert call_args
        assert call_args[0]
        assert call_args[0][0].startswith('Worker did not return running loops')


@pytest.mark.usefixtures('activated', 'watchdog_started_long_ago')
def test_took_too_long():
    with override_settings(
            {'worker': {'expected_http_response_time_seconds': 0}}):
        with httpretty.enabled():
            url = 'http://{}:{}/status/'.format(settings.get_worker_http_bind_address(),
                                                settings.get_worker_http_port())
            httpretty.register_uri(httpretty.GET, url, status=200, body='{}')
            httpretty.register_uri(httpretty.GET, settings.get_console_base_url(),
                                   status=200)  # emulate console running

            service = WatchdogService()

            service.reboot = MagicMock()
            service.restart_worker = MagicMock()
            service.start_worker = MagicMock()

            service.resolve_online_status = MagicMock(return_value=None)
            service.resolve_worker_supervisor_program_status = MagicMock(return_value=None)
            service.resolve_worker_process_status = MagicMock(return_value=None)
            service.resolve_loops_activity = MagicMock(return_value=None)
            service.resolve_workflow = MagicMock(return_value=None)
            service.report_error = MagicMock()
            service.report_warning = MagicMock()

            service.loop_iteration()

            service.start_worker.assert_not_called()
            service.restart_worker.assert_not_called()
            service.reboot.assert_not_called()
            service.resolve_loops_activity.assert_not_called()

            service.report_warning.assert_called()
            assert service.report_warning.call_args_list
            assert any(args[0][0].startswith('It took ') for args
                       in service.report_warning.call_args_list)
